// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.configuration;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;


/**
 * 
 * @author Neeraj
 * 
 * This class measures and displays time spent in a method and an entity (step or stream)
 *
 */
public class PerformanceAnalyzer 
{
	
	private HashMap _performanceMap = new HashMap();
	
	private String _entity;
	
	private HashMap _methodToStartTimeMap = new HashMap();
	
	private boolean _isEnabled = false;

	private boolean _enableDetailedPerf = false;
	
	/**
	 * Enables performance measurement for an entity (stream or step) based on properties
	 * passed via xJCL
	 * @param entityMonitored 
	 * @param props
	 */	
	
	public PerformanceAnalyzer(String entityMonitored, Properties props)
	{
		String enablePerfStr = props.getProperty(BDSFrameworkConstants.ENABLE_PERFORMANCE_MEASUREMENT_KEY);
		String enableDetailedPerfStr = props.getProperty(BDSFrameworkConstants.ENABLE_DETAILED_PERFORMANCE_MEASUREMENT_KEY);
		if(enablePerfStr != null && enablePerfStr.equals("true")) {
			_isEnabled = true;
		}	
		
		if(enableDetailedPerfStr != null && enableDetailedPerfStr.equals("true")) {
			_enableDetailedPerf  = true;
			_isEnabled = true;
		}
		
		_entity = entityMonitored;
	}
	public PerformanceAnalyzer(String entityMonitored) 
	{
		
		_entity = entityMonitored;
		
	}
	
	public boolean isEnabled()
	{
		return _isEnabled;
	}
	
	/**
	 * This method should be called on entering a method for which time spent needs
	 * to be measured.
	 * @param methodName
	 */
	public void startMeasurement(String methodName)
	{
		long startTime = System.currentTimeMillis();
		_methodToStartTimeMap.put(methodName,new Long(startTime));
	}
	
	/**
	 * This method should be called just before exiting a method for which time spent needs
	 * to be measured. The method will retrieve the previously saved begin time for the method
	 * and subtracts it from the current system time. The result is then saved using the addMeasurement
	 * method.
	 * @param methodName
	 */
	
	public void endMeasurementAndSave(String methodName)
	{
		long endTime = System.currentTimeMillis();
		long startTime = ((Long)_methodToStartTimeMap.get(methodName)).longValue();
		addMeasurement(methodName, endTime - startTime);
	}
	
	/**
	 * This method saves the time spent in a particular method. If an entry for this method
	 * already exists the given time is added to it else a new entry is created.
	 * @param methodName
	 * @param timeSpent
	 */
		
	public void addMeasurement(String methodName, long timeSpent) 
	{
		
		if(_performanceMap.containsKey(methodName)) 
		{
			PerformanceMetric pMetric = (PerformanceMetric)_performanceMap.get(methodName);
			
			pMetric.incrementMetric(timeSpent);
			
		}
		else
		{
			PerformanceMetric pMetric = new PerformanceMetric();
			pMetric.incrementMetric(timeSpent);
			_performanceMap.put(methodName, pMetric);
		}
	}
	
	/**
	 * This method displays various statistics about time spent in a particular method
	 * including total time and average time per call.
	 * @param methodName
	 */
	public void displayPerformanceStatisticsForMethod(String methodName)
	{
		System.out.println("****Performance statistics for method: " + methodName + " ****");
		if(_performanceMap.containsKey(methodName)) 
		{
			PerformanceMetric pMetric = (PerformanceMetric)_performanceMap.get(methodName);
			System.out.println("Total Time spent in method: " + pMetric.getTotalTime());
			System.out.println("Average Time spent per call in method: " + pMetric.getAverageTimePerCall());
		}
		else
		{
			System.out.println("No statistics for method: " + methodName);
		}
		System.out.println("******************************");
		
	}
	
	/**
	 * This method displays the performance statistics for the entity. It iterates over
	 * time spent in all methods displaying statistics for it and adding them to a cumulative
	 * total.
	 */
	public void displayPerformanceStatisticsForEntity()
	{
		System.out.println("****Performance statistics for Entity: " + _entity  + " ****");
		long totalTime = 0;
		Iterator perfIterator = _performanceMap.keySet().iterator();
		while(perfIterator.hasNext())
		{
			String methodName = (String)perfIterator.next();
			if (_enableDetailedPerf)
				displayPerformanceStatisticsForMethod(methodName);
			PerformanceMetric pMetric = (PerformanceMetric)_performanceMap.get(methodName);
			
			totalTime += pMetric.getTotalTime();
		}
		System.out.println("Total time spent in " + _entity + " is : " + totalTime);
		
		
	}
	
	private class PerformanceMetric 
	{
		private long _methodCallCount = 0;
		private long _cumulativeTime = 0;
		
		public PerformanceMetric() 
		{
			
		}
		
		public long getTotalTime() 
		{
			return _cumulativeTime;
		}
		
		public double getAverageTimePerCall()
		{
			if(_methodCallCount > 0)
			{	
				return _cumulativeTime / _methodCallCount;
			}
			else
				return 0;
		}
		
		public void incrementMetric(long timeSpent)
		{
			_cumulativeTime += timeSpent;
			++_methodCallCount;
		}
		
	}

}
