// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.datastreams.bdsadapter;


import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.BatchDataStream;

/**
 * Base implementation of {@link BatchDataStream}.
 * 
 * @author Antani / Seelbach / Neeraj
 */
public abstract class AbstractBatchDataStream implements BatchDataStream {

	protected String jobStepKey = "JobStepId";
       
	/** Job/step id. */
	private String jobstepId;

	/** BDS name. */
	private String name;

	/** Our properties, as specified in xJCL. */
	private Properties baseProperties;
	
	/** Are we restarting? */
	private Boolean restart;

	
	/**
	 * Close underlying implementation stream.
	 * 
	 * @throws BatchContainerDataStreamException
	 *   Underlying implementation stream threw an exception
	 *   in its <code>close()</code> method.
	 */
	
	public abstract void close() throws BatchContainerDataStreamException;    
	
	
	/**
	 * Parse restart token supplied by LREE.
	 * 
	 * <br>This implementation assumes the restart token
	 * is a record cound, parseable as an integer.
	 * 
	 * @throws RuntimeException
	 *   Restart token could not be parsed as an integer.
	 */
	public abstract void internalizeCheckpointInformation(String token) throws RuntimeException;

	/**
	 * Opens delegate stream.
	 * 
	 * @throws BatchContainerDataStreamException
	 *   Error opening delegate stream.
	 */
	public abstract void open() throws BatchContainerDataStreamException;

	/**
	 * Positions the stream to the initial checkpoint (i.e., the beginning
	 * of the stream).
	 * <br>This implementation simply sets the record counter to zero. 
	 * Overriding subclasses should call the parent implementation. 
	 * 
	 * @throws BatchContainerDataStreamException
	 *   Error positioning the stream.
	 */
	public abstract void positionAtInitialCheckpoint() throws BatchContainerDataStreamException;

	/**
	 * This method is invoked during stream initialization. Properties specified in the xJCL
	 * are passed as input
	 * @param props
	 */
	protected abstract void initialize(Properties props);
	
	/**
	 * This method is called just before a checkpoint is to be taken by the batch container.
	 * This allows the stream to save any information that can be used by the container in
	 * case of a restart.
	 */
	public abstract String externalizeCheckpointInformation();

	/**
	 * @return
	 *   Our job/step id.
	 */
	public String getJobstepId() {
		return jobstepId;
	}

	/**
	 * @return
	 *   Our logical name as specified in the xJCL.
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return
	 *   Our properties as specified in the xJCL.
	 */
	public Properties getProperties() {
		return this.baseProperties;
	}
    
	/**
	 * Convenience method to return a property value.
	 * @param key
	 *   Property key.
	 * @return
	 *   Property value.
	 */
    protected String getProperty(String key) {
        return getProperties().getProperty(key);
    }

	/**
	 * Convenience method to return a property value.
	 * Throws an exception if the property was not set, or was
	 * set to an empty value.
	 * 
	 * @param key
	 *   Property key.
	 * @return
	 *   Property value.
	 *  
	 * @throws RuntimeException
	 *   No value was specified for the given key.
	 */
    protected String getRequiredProperty(String key) {
        String value = getProperties().getProperty(key);
        if (value == null) {
            throw new RuntimeException("Missing required property '" + key + "'");
        }
        value = value.trim();
        if (value.length() == 0) {
            throw new RuntimeException("Required property '" + key + "' has empty value");
        }
        return value;
    }   
    
    protected String getRequiredProperty(Properties props,String key) {
        String value = props.getProperty(key);
        if (value == null) {
            throw new RuntimeException("Missing required property '" + key + "'");
        }
        value = value.trim();
        if (value.length() == 0) {
            throw new RuntimeException("Required property '" + key + "' has empty value");
        }
        return value;
    }   
    
    /**
     * Called during stream initialization.
     */
	public void initialize(String name, String jobstepId) throws BatchContainerDataStreamException {        
		this.name = name;
		this.jobstepId = jobstepId;

		try {
			
			this.baseProperties.put(jobStepKey, jobstepId);
			initialize(baseProperties);
		} catch (Exception e) {
			throw new BatchContainerDataStreamException(e);
		}
	}

	/**
	 * LREE calls this method after a checkpoint is taken
	 * 
	 * <br>This implementation does nothing; subclasses may override.
	 */
	public void intermediateCheckpoint() {		
	}
	
	
	/**
	 * LREE calls this method to tell us our properties
	 * as specified in the xJCL.
	 * 
	 * @param props
	 *   Stream properties from xJCL.
	 */
	public void setProperties(Properties props) {
		this.baseProperties = props;		
	}

	/**
	 * @return
	 * <ul>
	 *   <li><code>Boolean.FALSE</code> if we are not restarting
	 *   <li><code>Boolean.TRUE</code> if we are restarting
	 *   <li><code>null</code> if not yet known
	 * </ul>
	 */

	/**
	 * @return
	 * <ul>
	 *   <li><code>Boolean.FALSE</code> if we are not restarting
	 *   <li><code>Boolean.TRUE</code> if we are restarting
	 *   <li><code>null</code> if not yet known
	 * </ul>
	 */
	public Boolean isRestart() {
		return restart;
	}

	/**
	 * Sets the 'restarting' flag on this BDS. Should be called by
	 * subclasses, in their 
	 * {@link BatchDataStream#positionAtCurrentCheckpoint()} (set to <code>true</code>) or
	 * {@link BatchDataStream#positionAtInitialCheckpoint()} (set to <code>false</code>)
	 * methods, respectively.
	 * 
	 * @param restart
	 */
	protected void setRestart(boolean restart) {
		this.restart = Boolean.valueOf(restart);
	}
	
	protected Object loadClass(String className)
	{
		try
		{
			Object retval = Thread.currentThread().getContextClassLoader().loadClass(className).newInstance();
			return retval;
		}
		catch (Throwable t)
		{
			System.err.println("Exeption loading class " + className + t.getLocalizedMessage());
			throw new RuntimeException(t);
		}
			
	}
}
