// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

/**
 * @author Neeraj
 */

import java.sql.CallableStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

public abstract class BDSCallableStatementJDBCReader extends AbstractBatchDataStream 
implements AbstractBatchDataInputStream 
{
	public static final String CLASSNAME = BDSCallableStatementJDBCReader.class.getName();
	
	protected JDBCCallableStatementReaderPattern jdbcCallableStmtReaderImpl; 
	protected boolean prefetchedRecordHasBeenRead = false;
	protected Object record;
	protected ResultSet resultSet;
	protected CallableStatement cstmt = null;
	
	protected String restartToken;
	protected int recordCount;
	protected BDSFWLogger logger;
	protected IJDBCConnection jdbcConn;
	protected boolean holdCursors = false;
	
	
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;

	
	
	/**
	 * This method is called during stream initialization. All properties
	 * specified in the xJCL for this stream are provided as input to this method.
	 * This method will initialize the user's implementation of the 
	 * JDBCReaderPattern  
	 */
	public void initialize(Properties props)
	{
		try 
		{
			logger = new BDSFWLogger(props);
			perfAnalyzer = new PerformanceAnalyzer(CLASSNAME,props);
			isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
			if(jdbcConn == null)
			{	
				jdbcConn = new JDBCConnection();
				jdbcConn.initialize(props);
			}	
			
			String lpsEnabledStr = props.getProperty(BDSFrameworkConstants.LPS_ENABLED,"false");
			if(lpsEnabledStr.equals("true")) {
				logger.info("LPS enabled");
				holdCursors = true;
			}
			
			String holdCursorsEnabledStr = props.getProperty(BDSFrameworkConstants.CURSOR_HOLDABILITY_ENABLED,"false");
			if(holdCursorsEnabledStr.equals("true")) {
				logger.info("CursorHoldability enabled");
				holdCursors = true;
			}
			
			
			
			String jdbcReaderImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(jdbcReaderImplClass == null)
				jdbcReaderImplClass = getRequiredProperty(props,"IMPLCLASS");	
			jdbcCallableStmtReaderImpl = (JDBCCallableStatementReaderPattern) loadClass(jdbcReaderImplClass);
			if(logger.isDebugEnabled())
				logger.info("Loaded JDBCReaderPattern "  + jdbcReaderImplClass);
			jdbcCallableStmtReaderImpl.initialize(props);
			
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}
	
	/**
	 * Invokes the user's implementation of the JDBCCallableStmtReaderPattern getInitialCallableStatement()
	 * to obtain the initial lookup callable statement
	 * Next it executes the callable statement to obtain the resultSet
	 */
	

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {
		String methodName = "positionAtInitialCheckpoint";
		if(holdCursors) {
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.startMeasurement(methodName);
			
			try 
			{			
				recordCount = 0;
				cstmt = jdbcCallableStmtReaderImpl.getInitialCallableStatement(jdbcConn.getJDBCConnection());
				
				this.executeStatement(cstmt);
				
			}
			catch (Throwable t)
			{
				throw new BatchContainerDataStreamException(t);
			}		
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
		}	
	}
	
	protected void executeStatement(CallableStatement cstmt) {
		
		try {
				resultSet = cstmt.executeQuery();
			
		} catch (SQLException e) {
			logger.error("Error creating and executing statement");
			e.printStackTrace();
		}
	}
	
	/**
	 * Invokes the user's implementation of the JDBCCallableStmtReaderPattern getRestartCallableStatement()
	 * method to obtain restart callable statement
	 * and executes the callable statement to obtain the ResultSet
	 */
	public void positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {
		String methodName = "positionAtCurrentCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try 
		{
			cstmt = jdbcCallableStmtReaderImpl.getRestartCallableStatement(jdbcConn.getJDBCConnection(),restartToken);
			
			executeStatement(cstmt);

		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	/**
	 * Invokes the user's implementation of the JDBCCallableStatementReaderPattern 
	 * fetchRecord(resultSet) method to fetch
	 * the next record from the resultSet
	 * 
	 */
	public Object read() throws Exception {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		recordCount++;
		Object rec = jdbcCallableStmtReaderImpl.fetchRecord(resultSet); 
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return rec;
	}
	
	/**
	 * This method creates a Statement based on the SQL query passed
	 * executes the the statement and returns the resultSet
	 */
	public Object read(Object sql) throws Exception {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		Statement st =  jdbcConn.getJDBCConnection().createStatement();
		Object result  =  st.executeQuery((String)sql);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return result;
	}

	/**
	 * Checks whether the resultSet has more rows
	 */
	public boolean hasNext() throws Exception {
		if(resultSet != null )
			return resultSet.next();
		else {
			_positionAtInitialCheckpoint();
			return resultSet.next();
		}
			
	}
	
	private void _positionAtInitialCheckpoint() throws BatchContainerDataStreamException {
		String methodName = "positionAtInitialCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try 
		{			
			recordCount = 0;
			cstmt = jdbcCallableStmtReaderImpl.getInitialCallableStatement(jdbcConn.getJDBCConnection());
			
			
			this.executeStatement(cstmt);
			
			
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * Invoked during stream initialization. It is used to open
	 * a JDBC connection
	 */
	public void open() throws BatchContainerDataStreamException {
		String methodName = "open";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		jdbcConn.open();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	/**
	 * This method closes the resultSet and JDBC connection
	 */
	public void close() throws BatchContainerDataStreamException {

		try {		
			if(cstmt != null) {
				cstmt.close();
			}
			
			if(resultSet != null) {
				resultSet.close();
			}	
			jdbcConn.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();

		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}

	/**
	 * This method is invoked just before a checkpoint is taken.
	 * It invokes the user's implementation of the JDBCCallableStatementReaderPattern
	 * getRestartTokens() method to obtain the restartToken.
	 */
	public String externalizeCheckpointInformation() {
		String methodName = "externalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try
		{			
			restartToken = jdbcCallableStmtReaderImpl.getRestartTokens();	
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
			// if LPS is not enabled close everything on chkpt
			if(!holdCursors) {
				if(cstmt != null) {
					cstmt.close();
					cstmt = null;
				}
				
				if( resultSet != null) {
					resultSet.close();
					resultSet = null;
				}
				
			}	
			return restartToken;
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		
	}
	
	/**
	 * This method is invoked during a job restart. It invokes the 
	 * user's implementation of JDBCCallableStatementReaderPattern getRestartCallableStatement
	 * method with the token saved during the previous checkpoint. It then
	 * executes it to obtain the resultSet
	 */
	
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName + "- token= |" + token + "|");
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try 
		{
			
			restartToken = token;
			cstmt = jdbcCallableStmtReaderImpl.getRestartCallableStatement(
					jdbcConn.getJDBCConnection(),restartToken);
			
			
			this.executeStatement(cstmt);
		
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	/**
	 * Not implemented
	 */
	public Object fetchHeader() {
		// NOT APPLICABLE
		return null;
	}
	
	public void intermediateCheckpoint()  {		
		String methodName = "intermediateCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		if(!holdCursors) {		
			recordCount = 0;
			cstmt = jdbcCallableStmtReaderImpl.getRestartCallableStatement(jdbcConn.getJDBCConnection(),restartToken);
			
		
			this.executeStatement(cstmt);
				
			
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
		}	
	}



	
}
