//Licensed Materials - Property of IBM 
//5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.sql.CallableStatement;

import java.sql.SQLException;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

/**
* 
* @author Neeraj
* 
* This class is used by the JDBCCallableStatementWriterPattern to write data to JDBC resource
*
*/
public abstract class BDSCallableStatementJDBCWriter extends AbstractBatchDataStream 
	implements AbstractBatchDataOutputStream {
	
	public static final String CLASSNAME = BDSJDBCWriter.class.getName();
	protected JDBCCallableStatementWriterPattern jdbcCSWriterImpl;	
	
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;
	protected IJDBCConnection jdbcConn = null;
	
	protected int batchIncrementor = 0;
	protected int batcherInterval = 20; 
	protected CallableStatement cstmt = null;
	protected boolean dataWritten = false;
	protected boolean forceConnectionRecycle = false;
	
	/**
	 * Opens a JDBC Connection based on the parameters passed in the xJCL
	 */
	public void open() throws BatchContainerDataStreamException {
		if (!forceConnectionRecycle)
		jdbcConn.open();
	}
	
	/**
	 * Closes the JDBC Connection
	 */
	public void close() throws BatchContainerDataStreamException {
		if (!forceConnectionRecycle)
			jdbcConn.close();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.displayPerformanceStatisticsForEntity();
	}
	
	/**
	 * Invoked during step setup this method initializes the user's implementation 
	 * of the JDBCCallableStatementWriterPattern. Set auto-commit to false to enable batch updates
	 */
	public void initialize(Properties props)
	{
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		try 
		{
			if(jdbcConn == null)
			{	
				jdbcConn = new JDBCConnection();
				props.setProperty(BDSFrameworkConstants.AUTO_COMMIT_ENABLED_KEY, "false");
				jdbcConn.initialize(props);
			}	
			
			String prop = props.getProperty(BDSFrameworkConstants.BATCH_INTERVAL_KEY);
			if  (prop != null)
			{
				try 
				{
					batcherInterval = Integer.valueOf(prop).intValue();				
				}
				catch (Exception e)
				{
					System.out.println("Invalid JDBC Batching interval specified. Defaulting to: " + batcherInterval);
				}
			}
			String jdbcCSWriterImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(jdbcCSWriterImplClass == null)
				jdbcCSWriterImplClass = getRequiredProperty(props,"IMPLCLASS");		
			
			
			
			jdbcCSWriterImpl = (JDBCCallableStatementWriterPattern) loadClass(jdbcCSWriterImplClass);
			if(logger.isDebugEnabled())
				logger.info("Loaded jdbcCSWriterImpl:" + jdbcCSWriterImplClass);
			jdbcCSWriterImpl.initialize(props);
			String forceConnRecycle = props.getProperty(BDSFrameworkConstants.FORCE_CONNECTION_RECYCLE, "false");
			
			if(forceConnRecycle.equalsIgnoreCase("true")) {
				forceConnectionRecycle = true;
			}
			
			if (logger.isDebugEnabled())
				logger.info("BDCallableStatementJDBCWriter.initialize: setting forcing_connection_recycle to " + forceConnectionRecycle);
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}
	/**
	 * Not implemented
	 */
	public void flush() throws Exception {	
		// not sure what to do here...naybe force a stmt.execute? but not sure it's needed.
	}

	/**
	 * This method invokes the writeRecord(callableStatment, record) method on the 
	 * user's implementation of the JDBCCallableStatementWriterPattern. The updated callableStatment statement
	 * is then added to the callableStatment batch. The callableStatment  is executed
	 * if the batch interval has been reached.
	 */
	public void write(Object record) throws Exception {
		String methodName = "write";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		String method="write";
		if(logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + ":" + method);
		dataWritten = true;
		try {
			cstmt = jdbcCSWriterImpl.writeRecord(cstmt, record);
			cstmt.addBatch();
			
			batchIncrementor++;
			
			if ((batchIncrementor % batcherInterval) == 0)
			{
				cstmt.executeBatch();
				batchIncrementor = 0;
			}
		} catch ( SQLException sqle ) {
			StringBuffer allMsg = new StringBuffer();
			for ( SQLException nextSqle = sqle; nextSqle != null; nextSqle = nextSqle.getNextException() ){
				logger.error("batchExecute() exception: "+nextSqle.getLocalizedMessage());
				allMsg.append(nextSqle.getLocalizedMessage()+"\n");
			}
			sqle.printStackTrace();
			throw new RuntimeException(allMsg.toString());
		} catch(Exception th) {
			th.printStackTrace();
			throw th;
		}
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":" + method);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * If the recycleconnections is enabled this method will execute the
	 * prepared statement batch and close the prepared statement and jdbc connection
	 */
	public String externalizeCheckpointInformation() {
		String method = "externalizeCheckpointInformation";

		if(logger.isDebugEnabled())
			logger.debug("Enteringg " + CLASSNAME + ":" + method);
		try
		{	
			if (dataWritten)
			{
				// externalizeCheckpointInformation is driven during the initialization of the BDS.
				// therefore if no sql queries are executed during the init, there's no reason to 
				// drive executeBatch() here.
				if(cstmt != null) { //bug# 14278
					cstmt.executeBatch();	
				
					cstmt.close();
					cstmt = null;
				}	
				batchIncrementor = 0;
				
			}
		}
		catch ( SQLException sqle ) {
			StringBuffer allMsg = new StringBuffer();
			for ( SQLException nextSqle = sqle; nextSqle != null; nextSqle = nextSqle.getNextException() ){
				logger.error("batchExecute() exception: "+nextSqle.getLocalizedMessage());
				allMsg.append(nextSqle.getLocalizedMessage()+"\n");
			}
			sqle.printStackTrace();
			throw new RuntimeException(allMsg.toString());
		}
		catch (Throwable t)
		{
			
			t.printStackTrace();
			throw new RuntimeException(t);

		}
		if (this.forceConnectionRecycle) {
			try {
				if (logger.isDebugEnabled())
					logger.info("BDCallableStatementJDBCWriter.externalizeCheckpoint: forcing connection recycle, closing JDBC Connection");
					
					jdbcConn.close();
			} catch (Throwable t) {
				throw new RuntimeException(t);
			}
		}

		if(logger.isDebugEnabled())
		logger.debug("Exiting " + CLASSNAME + ":externalizeCheckpointInformation");
		return "";
	}
	
	/**
	 * This method is invoked after a checkpoint has been taken. This method will open a new jdbcconnection and create a callable
	 * statement using the query obtained by invoking getSQLQuery() on the user's implementation
	 * of the JDBCCallableStatementWriterPattern.
	 */
	public void intermediateCheckpoint() {	
		String methodName = "intermediateCheckpoint";
		if(logger.isDebugEnabled())
		   logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if (this.forceConnectionRecycle)
		{
			try {
				if (logger.isDebugEnabled())
					logger.info("BDCallableStatementJDBCWriter.intermediateCheckpoint: forcing connection recycle, opening JDBC Connection");
				jdbcConn.open();
			} catch (Throwable t) {
				throw new RuntimeException(t);
			}
		}

		String sql = jdbcCSWriterImpl.getCallableStatementSQL();
		try {
			cstmt = jdbcConn.getJDBCConnection().prepareCall(sql);
		} catch (SQLException e) {
			
			System.err.println(e.getMessage());
			e.printStackTrace();
		}	
		
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":" + methodName);
		
	}

	/**
	 * Not implemented
	 */
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		// not sure we need to do anything here. I dont think there is any 
		// repositioning logic needed for insert/update logic.
	}
	/**
	 * Not implemented
	 */
	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {		
	}
	/**
	 * Not implemented
	 */
	public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {
		// not sure there is anything to do here.
	}
	/**
	 * Not implemented
	 */
	public void writeHeader(Object record) throws Exception {
		// not applicable
		
	}

}
