// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
/**
 * @author Antani / Neeraj
 * This class is responsible for reading from a Byte stream
 */

package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

public abstract class BDSFileByteReader extends AbstractBatchDataStream 
	implements AbstractBatchDataInputStream {

	/**
	 * This class handles the basic book-keeping related to 
	 * reading a Byte stream and delegates the "business logic"
	 * to the implementer of ByteReaderPattern
	 */
	protected ByteReaderPattern byteReaderImpl; 
	
	/**
	 * Holds the byte input stream 
	 */
	protected BufferedInputStream bufferedByteReader; 
	
	protected boolean prefetchedRecordHasBeenRead = false;
	
	/**
	 * basic unit of processing
	 */
	protected Object record;
	
	/**
	 * Name of the file to be read
	 */
	protected String fileName = "";
	
	
	protected int recordCount;
	
	/**
	 * Holds the properties that the user provides in the xJCL
	 */
	protected Properties props;	
	
	protected BDSFWLogger logger;
	
	/**
	 * Holds the performance metrics for this class
	 */
	protected PerformanceAnalyzer perfAnalyzer;
	
	protected boolean isPerformanceMeasurementEnabled = false;
	
	private static final String  CLASSNAME = BDSFileByteReader.class.getName();
	
	public abstract void open() throws BatchContainerDataStreamException;
	
	/**
	 * Store the properties passed through the xJCL
	 * Instantiate the logger
	 * Instantiate the bytereader class that the user implements
	 * 
	 */
	public void initialize(Properties props)
	{
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		
		try 
		{	
			
			this.props = props;
			
			String implClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(implClass == null)
				implClass = getRequiredProperty(props,"IMPLCLASS");
			byteReaderImpl = (ByteReaderPattern) loadClass(implClass);
			if(logger.isDebugEnabled())
				logger.info("Succesfully initialized" + implClass);
			byteReaderImpl.initialize(props);				
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		
	}
	
	/**
	 * This method delegates retrieval of the next record to the user provided
	 * pattern implementation class
	 */
	public Object read() throws IOException {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		
		Object returnValue = null;
		recordCount++;		
		if (!prefetchedRecordHasBeenRead) {
			
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME + ": prefetched record hasn't been read. ");
			
			if (record != null) {
				prefetchedRecordHasBeenRead = true;
				
				if (logger.isDebugEnabled()) {
					logger.debug(CLASSNAME + ":prefetched record now read.");
					logger.debug("Exiting " + CLASSNAME + ":read");
				}
				returnValue = record;
			} else	{
				if (logger.isDebugEnabled()) {
					logger.debug(CLASSNAME + ": no prefetched record. Fetching it now.");
					logger.debug("Exiting " + CLASSNAME + ":read");
				}
				returnValue = byteReaderImpl.fetchRecord(bufferedByteReader);
			}
		} else {
			if (logger.isDebugEnabled())
				logger.debug("Exiting " + CLASSNAME + ":read");
			returnValue = byteReaderImpl.fetchRecord(bufferedByteReader);
		}	
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		
		return returnValue;
	}
	
	/**
	 * This method is not supported
	 */
	
	
	public Object read(Object parameters) throws Exception {
		throw new Exception("Unsupported");
		
	}
	
	/**
	 * Returns true if there is more data to be read
	 */

	public boolean hasNext() throws IOException {
		String methodName = "hasNext";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + ":" + methodName);
		
		boolean hsNxt = false;
		prefetchedRecordHasBeenRead = false;
		if (bufferedByteReader.available() != 0)
		{
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME + ": yes, hasNext. fetching record ");
			
			record = byteReaderImpl.fetchRecord(bufferedByteReader);
			hsNxt = true;
		}
		else
		{
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME + ": no more data in buffered byte reader.");
			hsNxt = false;
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		return hsNxt;
	}
	
	/**
	 * Close the stream
	 */

	public void close() throws BatchContainerDataStreamException {

		try {				
			bufferedByteReader.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}
	
	/**
	 * Return the current recordcount
	 */
	
	public String externalizeCheckpointInformation() {
		
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME  + ".externalizeCheckpointInfo- token= |" + recordCount + "|");
		
		return String.valueOf(recordCount);		
	}
	
	/**
	 * Position at the given record number
	 */
	
	public void internalizeCheckpointInformation(String token)
			throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		if (logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + "." + methodName +"- token= |" + token + "|");
		
		try {
			if (token == null)
				positionAtInitialCheckpoint();

			else
				recordCount = Integer.valueOf(token).intValue();
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if (logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + "." + methodName + "- token= |" + token + "|");
		
	}
	
	/**
	 * Invoked when the stream is first opened. 
	 */
	
	public void positionAtInitialCheckpoint()
			throws BatchContainerDataStreamException {
		recordCount = 0;
	}
	
	/**
	 * Start from the beginning and read until the current record = recordcount
	 */
	
	public void positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {			
		String methodName = "positionAtCurrentCheckpoint";
		if(logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		setRestart(true);
		int repositionCount = recordCount;
		recordCount = 0;
		try {			
			int i = 0;
			while ((i < repositionCount) && hasNext())
			{
				read();
				i++;
			}
			
		} catch (Exception e) {
			throw new BatchContainerDataStreamException(e);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":"  + methodName);
		
	}
	
	/**
	 * Invokes the user's implementation of the ByteReader pattern to
	 * retrieve the file header.
	 */
	public Object fetchHeader() {
		Object header = null;
		String methodName = "fetchHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		header = byteReaderImpl.fetchHeader();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return header;
	}
}
