// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008



package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

/**
 * @author Antani/Neeraj
 * This class is responsible for writing to a Byte stream
 */

import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

public abstract class BDSFileByteWriter extends AbstractBatchDataStream 
	implements AbstractBatchDataOutputStream {

	protected BufferedOutputStream writer;
	
	/**
	 * This class handles the basic book-keeping related to 
	 * writing to a Byte stream and delegates the "business logic"
	 * to the implementer of ByteReaderPattern
	 */
	protected ByteWriterPattern byteWriterPattern; 
	
	/**
	 * Current position of the stream
	 */
	protected long position = 0;
	
	protected boolean headerWritten = false;
	protected FileOutputStream fos;
	protected long currentPosition = 0;
	protected abstract long getCurrentPosition();
	protected abstract void setPosition(long position);
	public abstract void open() throws BatchContainerDataStreamException;
	protected BDSFWLogger logger;
	protected boolean isRestart = false;
	protected PerformanceAnalyzer perfAnalyzer;
	protected boolean isPerformanceMeasurementEnabled = false;
	
	public static final String CLASSNAME = BDSFileByteWriter.class.getName();
	public void flush() throws Exception {		
		writer.flush();		
	}

	/**
	 * Delegates writing of the passed record to the user's implementation
	 * of the FileByteWriter pattern.
	 */
	public void write(Object record) throws Exception {

		String methodName = "write";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + ": About to write record");
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		
		byteWriterPattern.writeRecord(writer, record);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + ": Wrote record");
	}

	/**
	 * Closes the outputstream.
	 */
	public void close() throws BatchContainerDataStreamException {
		try {			
			writer.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}	
	}

	/**
	 * Returns the current file pointer position
	 */
	public String externalizeCheckpointInformation() {
		String methodName = "externalizeCheckpointInformation";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + ": Entering " + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try{
			String checkpointToken = String.valueOf(getCurrentPosition());
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
			
			if (logger.isDebugEnabled())
				logger.debug("BDSFileByteWriter.externalizeCheckpointInfo- token= |" + checkpointToken + "|");
			
			return String.valueOf(getCurrentPosition());
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		
	}

	/**
	 * Invoked during stream initialization. This method loads the user provided
	 * pattern implementation class
	 */
	protected void initialize(Properties props) {
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		try
		{	
			String implClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(implClass == null)
				implClass = getRequiredProperty(props,"IMPLCLASS");
			byteWriterPattern = (ByteWriterPattern) loadClass(implClass);
			if (logger.isDebugEnabled())
				logger.info("Loaded " + implClass);
			byteWriterPattern.initialize(props);			
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}

	/**
	 * Invoked during a job restart. The the passed string is used in the subsequent
	 * call to positionAtCurrentCheckpoint to reposition the file pointer
	 */
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		
		String methodName = "internalizeCheckpointInformation";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + "." + methodName + "- token= |" + token + "|");
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		isRestart = true;
		position = Long.valueOf(token).longValue();		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * do nothing
	 */

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {				
	}

	/**
	 * 
	 * This method is invoked immediately after internalize checkpoint to allow 
	 * the stream to position itself
	 */
	public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {
		setPosition(position);
	} 
	
	/**
	 * Called during stream initialization to write header information if any.
	 */
	
	public void writeHeader(Object header) throws Exception {
		String methodName = "writeHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if(!isRestart)
			byteWriterPattern.writeHeader(writer, header);
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	
	
}
