// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

/**
 * @author Antani/Neeraj
 * This class is responsible for reading from a text stream
 */
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.FileReaderPattern;

public abstract class BDSFileReader extends AbstractBatchDataStream 
	implements AbstractBatchDataInputStream {

	public static final String CLASSNAME = "BDSFileReader";
	protected int recordCount;

	protected FileReaderPattern fileReaderImpl;

	protected boolean prefetchedRecordHasBeenRead = false;

	protected Object record;

	protected BufferedReader reader;

	protected String fileName = "";
	
	protected Properties props;	
	
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;
	protected BDSFWLogger logger;
	
	
	protected String fileEncoding = null;

	/**
	 * Invoked during stream initialization. This method loads the user provided
	 * pattern implementation class
	 */
	public void initialize(Properties props)
	{
		try 
		{
			logger = new BDSFWLogger(props);
			perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
			isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
			this.props = props;
			fileEncoding = props.getProperty(BDSFrameworkConstants.FILE_ENCODING_KEY);
			String fileReaderImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(fileReaderImplClass == null)
				fileReaderImplClass = getRequiredProperty(props,"IMPLCLASS");			
			fileReaderImpl = (FileReaderPattern) loadClass(fileReaderImplClass);
			if (logger.isDebugEnabled())
				logger.debug("Loaded filereaderpattern" + fileReaderImplClass);
			fileReaderImpl.initialize(props);
			
		}
		catch (Throwable t)
		{
			logger.error("Exception occured loading filereaderpattern" + t.getMessage());
			throw new RuntimeException(t);
		}
	}
	
	public abstract void open() throws BatchContainerDataStreamException;
	
	public void positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {
		String methodName = "positionAtCurrentCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		setRestart(true);
		int repositionCount = recordCount;
		recordCount = 0;
		try {
			for (int i = 0; i < repositionCount; i++) {
				read();
			}
		} catch (Exception e) {
			throw new BatchContainerDataStreamException(e);
		}
		if(isPerformanceMeasurementEnabled) 
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	public Object read() throws IOException {
		String methodName = "read";
		if (logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		Object returnValue = null;

		recordCount++;
		if (!prefetchedRecordHasBeenRead) {
			if (record != null) {
				prefetchedRecordHasBeenRead = true;
				
				returnValue = record;
			} else {
			
				returnValue = fileReaderImpl.fetchRecord(reader);
				
				
			}	
		} else {
			
			returnValue = fileReaderImpl.fetchRecord(reader);
		}	
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return returnValue;
	}
	
	
	public Object read(Object parameters) throws Exception {
		throw new Exception("Unsupported");
		
	}

	public boolean hasNext() throws IOException {
		String methodName = "hasNext";
		if(logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + "." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		prefetchedRecordHasBeenRead = false;
		boolean moreDataToRead = false;
		record = fileReaderImpl.fetchRecord(reader);
		if (record != null)
		{	
			moreDataToRead =  true;
		}	
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + "." + methodName + " with " + moreDataToRead);
		return moreDataToRead;
	}

	public void close() throws BatchContainerDataStreamException {

		try {
			reader.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}

	public String externalizeCheckpointInformation() {
		
		if (logger.isDebugEnabled())
			logger.debug("BDSFileReaderr.externalizeCheckpointInfo- token= |" + recordCount + "|");
		
		return String.valueOf(recordCount);
	}
	
	public void internalizeCheckpointInformation(String token)
			throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try {
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME +"." + methodName + "-  token= |" + token + "|");
			
			if (token == null)
				positionAtInitialCheckpoint();
			else			
				recordCount = Integer.parseInt(token);
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	public void positionAtInitialCheckpoint()
			throws BatchContainerDataStreamException {
		recordCount = 0;
	}
	
	public Object fetchHeader() {
		String methodName = "fetchHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		Object header = null;
		header = fileReaderImpl.fetchHeader();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return header;
	}
}
