// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008


/**
 * @author Antani/Neeraj
 * This class is responsible for writing to a text stream
 */
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.io.BufferedWriter;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

public abstract class BDSFileWriter extends AbstractBatchDataStream 
	implements AbstractBatchDataOutputStream {

	public static final String CLASSNAME = BDSFileWriter.class.getName();
	protected BufferedWriter writer;
	protected FileWriterPattern fileWriterImpl;
	protected boolean headerWritten = false;
	protected long currentPosition = 0;
	protected String fileEncoding = null;
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;
	protected boolean isRestart = false;
	
	/**
	 * This method is called during stream initialization. All properties
	 * specified in the xJCL for this stream are provided as input to this method.
	 * This method will initialize the user's implementation of the 
	 * FileWriterPattern  
	 */
	public void initialize(Properties props)
	{
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		try 
		{
			fileEncoding = props.getProperty(BDSFrameworkConstants.FILE_ENCODING_KEY);			
			String fileWriterImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(fileWriterImplClass == null)
				fileWriterImplClass = getRequiredProperty(props,"IMPLCLASS");		
			fileWriterImpl = (FileWriterPattern) loadClass(fileWriterImplClass);
			if(logger.isDebugEnabled())
				logger.debug("Loaded FileWriterPattern:" + fileWriterImplClass);
			fileWriterImpl.initialize(props);
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(logger.isDebugEnabled())
			logger.debug("Exiting: BDSFileWriter initialize");
	}
	
	/**
	 * This is an abstract method implemented by the TextFileWriter
	 * @return
	 */
	protected abstract long getCurrentPosition();
	
	/**
	 * This is an abstract method implemented by the TextFileWriter
	 * @param position
	 */
	protected abstract void setPosition(long position);
	
	
	public abstract void close() throws BatchContainerDataStreamException;
	public abstract void open() throws BatchContainerDataStreamException;
	
	public void flush() throws Exception {
		writer.flush();
	}

	/**
	 * Invokes the user's implementation of the FileWriterPattern to 
	 * write the record to the file
	 */
	public void write(Object record) throws Exception {
		String methodName = "write";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		fileWriterImpl.writeRecord(writer, record);
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * This method is invoked just before a checkpoint is taken to save
	 * the current file position
	 * Gets the current file pointer position and returns a string value 
	 * of it.
	 */
	public String externalizeCheckpointInformation() {	
		String methodName = "externalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		currentPosition = getCurrentPosition();
		
		if (logger.isDebugEnabled())
		{
			logger.debug(CLASSNAME + "." + methodName + " Current Position= " + currentPosition);
		}
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return String.valueOf(currentPosition);
	}

	/**
	 * This method is invoked during restart.
	 * Parses the given restart token into a long and sets it as the 
	 * current position.
	 */
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if (logger.isDebugEnabled()) {
			logger.debug(CLASSNAME + "." + methodName + ". Current Position= " + token);
		}
		
		
		if(isPerformanceMeasurementEnabled) {
			perfAnalyzer.startMeasurement(methodName);
		}
		isRestart = true;
		try 
		{
			if (token == null)
				positionAtInitialCheckpoint();
			else
			{	
				currentPosition = Long.valueOf(token).longValue();
			
			}	
		}
		catch(Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {
		currentPosition = 0;
	}

	/**
	 * This method is called after internalizeCheckpoint. It resets the file pointer
	 * to the position specified in the internalizeCheckpoint call
	 */
	public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {
		String methodName = "positionAtCurrentCheckpoint";
		if (logger.isDebugEnabled()) {
			logger.debug(CLASSNAME + "." + methodName);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		setPosition(currentPosition);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	/**
	 * Invokes the user's implementation of the TextWriterPattern to write header data
	 */
	public void writeHeader(Object header) throws Exception {
		String methodName = "writeHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if(!isRestart)
			fileWriterImpl.writeHeader(writer, header);
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
}
