// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

/**
 * @author Antani/Neeraj
 * This class is responsible for reading from JDBC resource
 */
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

public abstract class BDSJDBCReader extends AbstractBatchDataStream 
implements AbstractBatchDataInputStream 
{
	public static final String CLASSNAME = BDSJDBCReader.class.getName();
	
	protected JDBCReaderPattern jdbcReaderImpl; 
	protected boolean prefetchedRecordHasBeenRead = false;
	protected Object record;
	protected ResultSet resultSet;
	protected PreparedStatement pstmt = null;
	
	protected String restartToken;
	protected int recordCount;
	protected BDSFWLogger logger;
	protected IJDBCConnection jdbcConn;
	protected boolean holdCursors = false;
	
	
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;

	
	
	/**
	 * This method is called during stream initialization. All properties
	 * specified in the xJCL for this stream are provided as input to this method.
	 * This method will initialize the user's implementation of the 
	 * JDBCReaderPattern  
	 */
	public void initialize(Properties props)
	{
		try 
		{
			logger = new BDSFWLogger(props);
			perfAnalyzer = new PerformanceAnalyzer(CLASSNAME,props);
			isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
			if(jdbcConn == null)
			{	
				jdbcConn = new JDBCConnection();
				jdbcConn.initialize(props);
			}	
			
			String lpsEnabledStr = props.getProperty(BDSFrameworkConstants.LPS_ENABLED,"false");
			if(lpsEnabledStr.equals("true")) {
				logger.info("LPS enabled");
				holdCursors = true;
			}
			
			String holdCursorsEnabledStr = props.getProperty(BDSFrameworkConstants.CURSOR_HOLDABILITY_ENABLED,"false");
			if(holdCursorsEnabledStr.equals("true")) {
				logger.info("CursorHoldability enabled");
				holdCursors = true;
			}
			
			
			
			String jdbcReaderImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(jdbcReaderImplClass == null)
				jdbcReaderImplClass = getRequiredProperty(props,"IMPLCLASS");	
			jdbcReaderImpl = (JDBCReaderPattern) loadClass(jdbcReaderImplClass);
			if(logger.isDebugEnabled())
				logger.info("Loaded JDBCReaderPattern "  + jdbcReaderImplClass);
			jdbcReaderImpl.initialize(props);
			
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}
	
	/**
	 * Invokes the user's implementation of the JDBCReaderPattern getInitialLookupQuery()
	 * to obtain the initial lookup query
	 * Next it builds a prepardstatement based on this query
	 * and executes it to obtain the resultSet
	 */
	

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {
		String methodName = "positionAtInitialCheckpoint";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(holdCursors) {
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.startMeasurement(methodName);
			
			try 
			{			
				recordCount = 0;
				String query = jdbcReaderImpl.getInitialLookupQuery();
				/*PreparedStatement pstmt = null;
				try {
					pstmt = jdbcConn.getJDBCConnection().prepareStatement(query);
					resultSet = pstmt.executeQuery();
				} catch (SQLException e) {
					logger.error("Error positioning at initial checkpoint");
					e.printStackTrace();
				}*/
				this.createStmtAndExecute(query);
				
			}
			catch (Throwable t)
			{
				throw new BatchContainerDataStreamException(t);
			}		
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
		}	
	}
	
	protected void createStmtAndExecute(String query) throws SQLException {
		String methodName = "createStmtAndExecute";
		
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName + " query: " + query);
		Connection c = jdbcConn.getJDBCConnection();
		
		pstmt = c.prepareStatement(query);
		if (logger.isDebugEnabled())
			logger.debug("created prepared statement: " + pstmt);
		resultSet = pstmt.executeQuery();
		if (logger.isDebugEnabled())
			logger.debug("executed prepared statement: ");
		
			
		
	}
	
	/**
	 * Invokes the user's implementation of the JDBCReaderPattern getRestartQuery(restartToken)
	 * method to obtain restartquery
	 * Next it creates a prepared statement and executes the 
	 * query to obtain the ResultSet
	 */
	public void positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {
		String methodName = "positionAtCurrentCheckpoint";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try 
		{
			String query = jdbcReaderImpl.getRestartQuery(restartToken);
			
			createStmtAndExecute(query);

		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	/**
	 * Invokes the user's implementation of the JDBCReaderPattern 
	 * fetchRecord(resultSet) method to fetch
	 * the next record from the resultSet
	 * 
	 */
	public Object read() throws Exception {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		recordCount++;
		Object rec = jdbcReaderImpl.fetchRecord(resultSet); 
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return rec;
	}
	
	/**
	 * This method creates a Statement based on the SQL query passed
	 * executes the the statement and returns the resultSet
	 */
	public Object read(Object sql) throws Exception {
		String methodName = "read";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		Statement st =  jdbcConn.getJDBCConnection().createStatement();
		Object result  =  st.executeQuery((String)sql);
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return result;
	}

	/**
	 * Checks whether the resultSet has more rows
	 */
	public boolean hasNext() throws Exception {
		boolean hasNxt = false;
		if(resultSet != null ) {
			hasNxt = resultSet.next();
			
			return hasNxt;
		}	
		else {
			_positionAtInitialCheckpoint();
			hasNxt= resultSet.next();
			
			return hasNxt;
		}
			
	}
	
	private void _positionAtInitialCheckpoint() throws BatchContainerDataStreamException {
		String methodName = "positionAtInitialCheckpoint";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try 
		{			
			recordCount = 0;
			String query = jdbcReaderImpl.getInitialLookupQuery();
			
			/*try {
				pstmt = jdbcConn.getJDBCConnection().prepareStatement(query);
				resultSet = pstmt.executeQuery();
			} catch (SQLException e) {
				logger.error("Error positioning at initial checkpoint");
				e.printStackTrace();
			}*/
			this.createStmtAndExecute(query);
			
			
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * Invoked during stream initialization. It is used to open
	 * a JDBC connection
	 */
	public void open() throws BatchContainerDataStreamException {
		String methodName = "open";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		jdbcConn.open();
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	/**
	 * This method closes the resultSet and JDBC connection
	 */
	public void close() throws BatchContainerDataStreamException {
		String methodName = "close";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		try {		
			if(pstmt != null) {
				pstmt.close();
			}
			
			if(resultSet != null) {
				resultSet.close();
			}	
		
			jdbcConn.close();
			
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();

		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}

	/**
	 * This method is invoked just before a checkpoint is taken.
	 * It invokes the user's implementation of the JDBCReaderPattern
	 * getRestartTokens() method to obtain the restartToken.
	 */
	public String externalizeCheckpointInformation() {
		String methodName = "externalizeCheckpointInformation";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try
		{			
			restartToken = jdbcReaderImpl.getRestartTokens();	
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
			// if LPS is not enabled close everything on chkpt
			if(!holdCursors) {
				if(pstmt != null) {
					pstmt.close();
					pstmt = null;
				}
				
				if( resultSet != null) {
					resultSet.close();
					resultSet = null;
				}
				
			}	
			return restartToken;
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		
	}
	
	/**
	 * This method is invoked during a job restart. It invokes the 
	 * user's implementation of JDBCReaderPattern getRestartQuery(token)
	 * method with the token saved during the previous checkpoint. It then
	 * creates a prepared statement and executes it to obtain the resultSet
	 */
	
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName + "- token= |" + token + "|");
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try 
		{
			
			restartToken = token;
			String query = jdbcReaderImpl.getRestartQuery(token);
			
			/*
			try {
				pstmt = jdbcConn.getJDBCConnection().prepareStatement(query);
				resultSet = pstmt.executeQuery();
			} catch (SQLException e) {
				logger.error("Exception occured while internalizecheckpoint");
				e.printStackTrace();
			}*/
			this.createStmtAndExecute(query);
		
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	/**
	 * Not implemented
	 */
	public Object fetchHeader() {
		// NOT APPLICABLE
		return null;
	}
	
	public void intermediateCheckpoint()  {		
		String methodName = "intermediateCheckpoint";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME +"." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		if(!holdCursors) {		
			recordCount = 0;
			String query = jdbcReaderImpl.getRestartQuery(restartToken);
			
			/*try {
				pstmt = jdbcConn.getJDBCConnection().prepareStatement(query);
				resultSet = pstmt.executeQuery();
			} catch (SQLException e) {
				logger.error("Error positioning datastream");
				e.printStackTrace();
			}*/
			try {
				this.createStmtAndExecute(query);
			} catch (Throwable e) {
				throw new RuntimeException(e);
			}
				
			
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
		}	
	}



	
}