// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patternadapter;


import java.util.Properties;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityTransaction;
import javax.persistence.Persistence;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;

public abstract class BDSJPAWriter extends AbstractBatchDataStream 
	implements AbstractBatchDataOutputStream {
	
	public static final String CLASSNAME = BDSJPAWriter.class.getName();
	protected JPAWriterPattern jpaWriterImpl;	
	
	private EntityManagerFactory _entityManagerFactory = null;
	private EntityManager _entityManager = null;
	
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;
	private boolean useJTATransactions = true;
	
	
	public void open() throws BatchContainerDataStreamException {
		_entityManager = this._entityManagerFactory.createEntityManager();
	}
	
	public void close() throws BatchContainerDataStreamException {
		try {			
			_entityManager.close();
			_entityManagerFactory.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();

		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}
	
	public void initialize(Properties props)
	{
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		try 
		{
			String persistentUnit = 
				getRequiredProperty(props,BDSFrameworkConstants.PERSISTENT_UNIT_KEY);
			
			if(_entityManagerFactory == null)
			{	
				_entityManagerFactory = 
					Persistence.createEntityManagerFactory(persistentUnit, props);
			}	
			String jpaWriterImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(jpaWriterImplClass == null)
				jpaWriterImplClass = getRequiredProperty(props,"IMPLCLASS");		
			
			String useJTATransactionsStr = props.getProperty(BDSFrameworkConstants.USE_JTA_TRAN,"true");
			if(useJTATransactionsStr.equals("false")) {
				useJTATransactions  = false;
			}
			
			
			jpaWriterImpl = (JPAWriterPattern) loadClass(jpaWriterImplClass);
			if(logger.isDebugEnabled())
				logger.info("Loaded JPAWriterPattern:" + jpaWriterImplClass);
			jpaWriterImpl.initialize(props);
			
			
			if (logger.isDebugEnabled())
				logger.info("BDSJPAWriter.initialize: ");
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}
	
	public void flush() throws Exception {	
		// not sure what to do here...naybe force a stmt.execute? but not sure it's needed.
	}

	public void write(Object record) throws Exception {
		String methodName = "write";
		if(logger.isDebugEnabled()) 
			logger.debug("Entering " + CLASSNAME + ":" + methodName + "rec:" + record);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if(!useJTATransactions) {
			if(_entityManager.isOpen()) {
				if(!_entityManager.getTransaction().isActive()) {
					_entityManager.getTransaction().begin();
				}	
			}	
		}
        // Create and persist a new Message entity
		try {
			jpaWriterImpl.writeRecord(_entityManager, record);
		} catch (Throwable th) {
			//_entityManager.getTransaction().rollback();
			throw new Exception(th.getCause());
		}
						
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":" + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	public String externalizeCheckpointInformation() {
		String methodName = "externalizeCheckpointInformation";
		if(logger.isDebugEnabled()) 
			logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if(!useJTATransactions) {
			if(_entityManager.isOpen()) {
				
				EntityTransaction et = _entityManager.getTransaction();
				if(et.isActive()) {
					et.commit();
				}
			}
		}	
		if(logger.isDebugEnabled()) 
			logger.debug("Exiting " + CLASSNAME + ":" + methodName);
		return "";
	}
	
	public void intermediateCheckpoint() {	
		String methodName = "intermediateCheckpoint";
		if(logger.isDebugEnabled()) 
			logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if(_entityManager.isOpen()) {
			if(!useJTATransactions) {
				_entityManager.getTransaction().begin();
			} else {	
				_entityManager.joinTransaction();
			}
		}	
		if(logger.isDebugEnabled()) 
			logger.debug("Exiting " + CLASSNAME + ":" + methodName);
	}

	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		// not sure we need to do anything here. I dont think there is any 
		// repositioning logic needed for insert/update logic.
	}

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {		
	}

	public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {
		// not sure there is anything to do here.
	}
	
	public void writeHeader(Object record) throws Exception {
		// not applicable
		
	}

}
