// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import com.ibm.pdq.runtime.Data;
import com.ibm.pdq.runtime.ResultIterator;
import com.ibm.pdq.runtime.factory.DataFactory;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.IJDBCConnection;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.JDBCConnection;
import java.sql.Connection;
import java.sql.Statement;
import java.util.Iterator;
import java.util.Properties;



public abstract class BDSPureQueryReader extends AbstractBatchDataStream
    implements AbstractBatchDataInputStream
{

	public static final String CLASSNAME = BDSPureQueryReader.class.getSimpleName();
    protected PureQueryReaderPattern pureQueryReaderImpl;
    protected Data pqDataInterface;
    protected String pqDataInterfaceStr;
    protected Class dataInterfaceClass;
    protected String dbSchema;
    protected boolean prefetchedRecordHasBeenRead;
    protected Object record;
    protected Iterator pqDataIterator;
    protected String restartToken;
    protected int recordCount;
    protected BDSFWLogger logger;
    protected IJDBCConnection jdbcConn;
    protected PerformanceAnalyzer perfAnalyzer;
    protected boolean isPerformanceMeasurementEnabled;
    public BDSPureQueryReader()
    {
        prefetchedRecordHasBeenRead = false;
        perfAnalyzer = null;
        isPerformanceMeasurementEnabled = false;
    }

    public void initialize(Properties properties)
    {
        try
        {
            logger = new BDSFWLogger(properties);
            logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering initialize()").toString());
            perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, properties);
            isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
            if(jdbcConn == null)
            {
                jdbcConn = new JDBCConnection();
                jdbcConn.initialize(properties);
                logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": initialized").toString());
            }
            String s = properties.getProperty("PATTERN_IMPL_CLASS");
            if(s == null)
                s = getRequiredProperty(properties, "IMPLCLASS");
            pureQueryReaderImpl = (PureQueryReaderPattern)loadClass(s);
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": Loaded PureQueryReaderPattern ").append(s).toString());
            dbSchema = properties.getProperty("DB_SCHEMA");
            if(dbSchema != null)
                dbSchema = dbSchema.trim();
            pqDataInterfaceStr = properties.getProperty("PQ_DATA_BEAN_INTERFACE");
            if(pqDataInterfaceStr != null)
            {
                pqDataInterfaceStr = pqDataInterfaceStr.trim();
                dataInterfaceClass = Thread.currentThread().getContextClassLoader().loadClass(pqDataInterfaceStr);
            }
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": User given pureQuery data interfaceloaded: ").append(pqDataInterfaceStr).toString());
            pureQueryReaderImpl.initialize(properties);
            logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting initialize()").toString());
        }
        catch(Throwable throwable)
        {
            throw new RuntimeException(throwable);
        }
    }

    public void positionAtInitialCheckpoint()
        throws BatchContainerDataStreamException
    {
        String s = "positionAtInitialCheckpoint";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering positionAtInitialCheckpoint()").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        try
        {
            if(dataInterfaceClass != null)
                pqDataInterface = (Data)DataFactory.getData(dataInterfaceClass, jdbcConn.getJDBCConnection());
            else
                pqDataInterface = DataFactory.getData(jdbcConn.getJDBCConnection());
            logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": Data interface obtained from purequery DataFactory").toString());
            setUserGivenSchema();
            recordCount = 0;
            pqDataIterator = pureQueryReaderImpl.getInitialIterator(pqDataInterface);
        }
        catch(Throwable throwable)
        {
            throw new BatchContainerDataStreamException(throwable);
        }
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting positionAtInitialCheckpoint()").toString());
    }

    private void setUserGivenSchema()
        throws Exception
    {
        String s = "setUserGivenSchema";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering setUserGivenSchema()").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        try
        {
            if(dbSchema != null && dbSchema.length() > 0)
            {
                Statement statement = jdbcConn.getJDBCConnection().createStatement();
                statement.execute((new StringBuilder("set schema='")).append(dbSchema).append("'").toString());
                statement.close();
                if(logger.isDebugEnabled())
                    logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": User given database schema: ").append(dbSchema).toString());
            }
        }
        catch(Throwable throwable)
        {
            throw new BatchContainerDataStreamException(throwable);
        }
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting setUserGivenSchema()").toString());
    }

    public void positionAtCurrentCheckpoint()
        throws BatchContainerDataStreamException
    {
        String s = "positionAtCurrentCheckpoint";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering positionAtCurrentCheckpoint()").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        try
        {
            if(dataInterfaceClass != null)
                pqDataInterface = (Data)DataFactory.getData(dataInterfaceClass, jdbcConn.getJDBCConnection());
            else
                pqDataInterface = DataFactory.getData(jdbcConn.getJDBCConnection());
            logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": Data interface obtained from purequery DataFactory, after restart").toString());
            setUserGivenSchema();
            pqDataIterator = pureQueryReaderImpl.getRestartIterator(pqDataInterface, restartToken);
        }
        catch(Throwable throwable)
        {
            throw new BatchContainerDataStreamException(throwable);
        }
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting positionAtCurrentCheckpoint()").toString());
    }

    public Object read()
        throws Exception
    {
        String s = "read";
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering read()").toString());
        recordCount++;
        Object obj = pureQueryReaderImpl.fetchRecord(pqDataIterator);
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting read()").toString());
        return obj;
    }

    public Object read(Object obj)
        throws Exception
    {
        /*String s = "read";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering read(Object id)").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        Object obj1 = pureQueryReaderImpl.fetchRecord(pqDataIterator, obj);
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting read(Object id)").toString());
        return obj1;*/
    	throw new Exception("Not Supported");
    }

    public boolean hasNext()
        throws Exception
    {
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering hasNext()").toString());
        boolean flag;
        if(pqDataIterator != null)
        {
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": calling pureQuery iterator hasNext() ").toString());
            flag = pqDataIterator.hasNext();
        } else  {
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": iterator was closed; will position at initial point").toString());
            if(dataInterfaceClass != null)
                pqDataInterface = (Data)DataFactory.getData(dataInterfaceClass, jdbcConn.getJDBCConnection());
            else
                pqDataInterface = DataFactory.getData(jdbcConn.getJDBCConnection());
            logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": Data interface obtained from purequery DataFactory").toString());
            setUserGivenSchema();
            recordCount = 0;
            pqDataIterator = pureQueryReaderImpl.getInitialIterator(pqDataInterface);
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": positioned at initial point; calling pureQuery iterator hasNext() ").toString());
            flag = pqDataIterator.hasNext();
        }
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting hasNext()").toString());
        return flag;
    }

    public void open()
        throws BatchContainerDataStreamException
    {
        String s = "open";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering open()").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        jdbcConn.open();
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting open()").toString());
    }

    public void close()
        throws BatchContainerDataStreamException
    {
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering close()").toString());
        try
        {
            if(pqDataIterator != null)
                ((ResultIterator)pqDataIterator).close();
            if(pqDataInterface != null)
                pqDataInterface.close();
            jdbcConn.close();
            if(isPerformanceMeasurementEnabled)
                perfAnalyzer.displayPerformanceStatisticsForEntity();
        }
        catch(Throwable throwable)
        {
            throw new BatchContainerDataStreamException(throwable);
        }
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting close()").toString());
    }

    public String externalizeCheckpointInformation()
    {
        String s = "externalizeCheckpointInformation";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering externalizeCheckpointInformation()").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        try
        {
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": getting restart token ").toString());
            restartToken = pureQueryReaderImpl.getRestartTokens();
            if(pqDataIterator != null)
            {
                ((ResultIterator)pqDataIterator).close();
                pqDataIterator = null;
                if(logger.isDebugEnabled())
                    logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": closed iterator ").toString());
            }
            if(pqDataInterface != null)
            {
                pqDataInterface.close();
                pqDataInterface = null;
                if(logger.isDebugEnabled())
                    logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": closed data interface ").toString());
            }
            if(jdbcConn != null)
            {
                jdbcConn.close();
                if(logger.isDebugEnabled())
                    logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": closed jdbc connection ").toString());
            }
        }
        catch(Throwable throwable)
        {
            throw new RuntimeException(throwable);
        }
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting externalizeCheckpointInformation()").toString());
        return restartToken;
    }

    public void intermediateCheckpoint()
    {
        String s = "intermediateCheckpoint";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering intermediateCheckpoint()").toString());
        if(logger.isDebugEnabled())
            logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(".").append(s).toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s);
        try
        {
            jdbcConn.open();
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": reopened connection ").toString());
            if(dataInterfaceClass != null)
                pqDataInterface = (Data)DataFactory.getData(dataInterfaceClass, jdbcConn.getJDBCConnection());
            else
                pqDataInterface = DataFactory.getData(jdbcConn.getJDBCConnection());
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": got data interface from pureQuery ").toString());
            setUserGivenSchema();
            if(logger.isDebugEnabled())
                logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(": getting restart iterator from application ").toString());
            pqDataIterator = pureQueryReaderImpl.getRestartIterator(pqDataInterface, restartToken);
        }
        catch(Throwable throwable)
        {
            throw new RuntimeException(throwable);
        }
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting intermediateCheckpoint()").toString());
    }

    public void internalizeCheckpointInformation(String s)
        throws RuntimeException
    {
        String s1 = "internalizeCheckpointInformation";
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": entering internalizeCheckpointInformation()").toString());
        if(logger.isDebugEnabled())
            logger.debug((new StringBuilder(String.valueOf(CLASSNAME))).append(".").append(s1).append("- token= |").append(s).append("|").toString());
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.startMeasurement(s1);
        try
        {
            restartToken = s;
            if(dataInterfaceClass != null)
                pqDataInterface = (Data)DataFactory.getData(dataInterfaceClass, jdbcConn.getJDBCConnection());
            else
                pqDataInterface = DataFactory.getData(jdbcConn.getJDBCConnection());
            setUserGivenSchema();
            pqDataIterator = pureQueryReaderImpl.getRestartIterator(pqDataInterface, s);
        }
        catch(Throwable throwable)
        {
            throw new RuntimeException(throwable);
        }
        if(isPerformanceMeasurementEnabled)
            perfAnalyzer.endMeasurementAndSave(s1);
        logger.info((new StringBuilder(String.valueOf(CLASSNAME))).append(": exiting internalizeCheckpointInformation()").toString());
    }

    public Object fetchHeader()
    {
        return null;
    }

    

}
