// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.sql.Statement;
import java.util.Properties;

import com.ibm.pdq.runtime.Data;
import com.ibm.pdq.runtime.HeterogeneousBatchKind;
import com.ibm.pdq.runtime.factory.DataFactory;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.IJDBCConnection;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.JDBCConnection;



/**
 * 
 * @author Neeraj
 * 
 * This method is used by the JDBCWriterPattern to write data to JDBC resource
 *
 */
public abstract class BDSPureQueryWriter extends AbstractBatchDataStream 
	implements AbstractBatchDataOutputStream {
	
	public static final String CLASSNAME = BDSPureQueryWriter.class.getName();
	protected PureQueryWriterPattern jdbcWriterImpl;	
	
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;
	protected IJDBCConnection jdbcConn = null;
	protected Data pqDataInterface;
	protected String pqDataInterfaceStr;
	protected Class dataInterfaceClass;
	protected String dbSchema;
	
	protected int batchIncrementor = 0;
	protected int batcherInterval = 20; 
	protected boolean dataWritten = false;
	boolean forceConnectionRecycle = false;
	
	/**
	 * Opens a JDBC Connection based on the parameters passed in the xJCL
	 */
	public void open() throws BatchContainerDataStreamException {
		if (!forceConnectionRecycle) {
			jdbcConn.open();
		}
	}
	
	/**
	 * Closes the JDBC Connection
	 */
	public void close() throws BatchContainerDataStreamException {
		if (!forceConnectionRecycle) {
			if ( pqDataInterface != null ) {
				pqDataInterface.close();
			}
			jdbcConn.close();
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.displayPerformanceStatisticsForEntity();
	}
	
	/**
	 * Invoked during step setup this method initializes the user's implementation 
	 * of the TiTaPureQueryWriterPattern. Set auto-commit to false to enable batch updates
	 */
	public void initialize(Properties props)
	{
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		try 
		{
			if(jdbcConn == null)
			{	
				jdbcConn = new JDBCConnection();
				props.setProperty(BDSFrameworkConstants.AUTO_COMMIT_ENABLED_KEY, "false");
				jdbcConn.initialize(props);
			}	
			
			String prop = props.getProperty(BDSFrameworkConstants.BATCH_INTERVAL_KEY);
			if  (prop != null)
			{
				try 
				{
					batcherInterval = Integer.valueOf(prop).intValue();				
				}
				catch (Exception e)
				{
					System.out.println("Invalid JDBC Batching interval specified. Defaulting to: " + batcherInterval);
				}
			}
			String jdbcWriterImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(jdbcWriterImplClass == null)
				jdbcWriterImplClass = getRequiredProperty(props,"IMPLCLASS");					
			
			jdbcWriterImpl = (PureQueryWriterPattern) loadClass(jdbcWriterImplClass);
			if(logger.isDebugEnabled())
				logger.info("Loaded PureQueryWriterPattern:" + jdbcWriterImplClass);
			jdbcWriterImpl.initialize(props);
			String forceConnRecycle = props.getProperty(BDSFrameworkConstants.FORCE_CONNECTION_RECYCLE, "false");
			
			if(forceConnRecycle.equalsIgnoreCase("true")) {
				forceConnectionRecycle = true;
			}			
			if (logger.isDebugEnabled())
				logger.info("BDSPureQueryWriter.initialize: setting forcing_connection_recycle to " + forceConnectionRecycle);
			
			// get the user supplied db schema, if any
			dbSchema = props.getProperty( BDSFrameworkConstants.DB_SCHEMA_KEY );
			if ( dbSchema != null )  {
				dbSchema = dbSchema.trim();
				if (logger.isDebugEnabled())
					logger.info("BDSPureQueryWriter.initialize: User DB schema: " + dbSchema);
			}
			
			// get user given pureQuery data interface, if any
			pqDataInterfaceStr = props.getProperty( BDSFrameworkConstants.PQ_DATA_INTERFACE_KEY );
			if ( pqDataInterfaceStr != null ) {
				pqDataInterfaceStr = pqDataInterfaceStr.trim();
				if (logger.isDebugEnabled())
					logger.info("BDSPureQueryWriter.initialize: User Data Interface class: " + pqDataInterfaceStr);
				dataInterfaceClass = Thread.currentThread().getContextClassLoader().loadClass( pqDataInterfaceStr );
				if (logger.isDebugEnabled())
					logger.info("BDSPureQueryWriter.initialize: Class " + pqDataInterfaceStr + " loaded");
			}
			
			if ( logger.isDebugEnabled() ) {
				logger.debug( CLASSNAME + ": User given pureQuery data interfaceloaded: " + pqDataInterfaceStr );
			}
			
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}	
	
	private void setUserGivenSchema() throws Exception {
		String methodName = "setUserGivenSchema";
		logger.info( CLASSNAME + ": entering setUserGivenSchema()" );
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try {			
			// if schema given, set schema
			if ( ( dbSchema != null ) && ( dbSchema.length() > 0 ) ) {
				Statement stmt = jdbcConn.getJDBCConnection().createStatement();
				stmt.execute( "set schema='" + dbSchema + "'" );
				stmt.close();
				if ( logger.isDebugEnabled() ) {
					logger.info( CLASSNAME + ": User given database schema: " + dbSchema );
				}
			}
		}
		catch ( Throwable t ) {
			throw new BatchContainerDataStreamException( t ); 
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		logger.info( CLASSNAME + ": exiting setUserGivenSchema()" );
	}

	/**
	 * Not implemented
	 */
	public void flush() throws Exception {	
		// not sure what to do here...naybe force a stmt.execute? but not sure it's needed.
	}

	/**
	 * This method invokes the writeRecord(Data, record) method on the 
	 * user's implementation of the TiTaPureQueryWriterPattern. The updated prepared statement
	 * is then added to the preparedstatment batch. The prepared statement is executed
	 * if the batch interval has been reached.
	 */
	public void write(Object record) throws Exception {
		String methodName = "write";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		String method="write";
		if(logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + ":" + method);
		
		if ( batchIncrementor == 0 ) {
			pqDataInterface.startBatch( HeterogeneousBatchKind.heterogeneousModify__);
		}
		
		dataWritten = true;
		
		jdbcWriterImpl.writeRecord(pqDataInterface, record);
		
		batchIncrementor++;
		
		if ((batchIncrementor % batcherInterval) == 0)
		{
			// specified batch interval exceeded. So, force pureQuery to do executeBatch() and 
			// start a new batch
			pqDataInterface.endBatch();
			batchIncrementor = 0;
		}
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":" + method);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * If the recycleconnections is enabled this method will execute the
	 * prepared statement batch and close the prepared statement and jdbc connection
	 */
	public String externalizeCheckpointInformation() {
		String method = "externalizeCheckpointInformation";

		if(logger.isDebugEnabled())
			logger.debug("Enteringg " + CLASSNAME + ":" + method);
		try
		{	
			if (dataWritten)
			{
				if( (pqDataInterface != null) && ( batchIncrementor > 0 ) ) { 
					pqDataInterface.endBatch();
				}	
				batchIncrementor = 0;
				
			}
			else { // this is the first dummy checkpointing soon after BDS initialization
				if ( !forceConnectionRecycle ) {
					// get Data implementation from pureQuery
					if ( dataInterfaceClass != null )
						pqDataInterface = (Data) ( DataFactory.getData( dataInterfaceClass, jdbcConn.getJDBCConnection() ) );
					else 
						pqDataInterface = (Data) ( DataFactory.getData( jdbcConn.getJDBCConnection() ) );
					if (logger.isDebugEnabled()) {
						logger.info( CLASSNAME + ": Data interface obtained from purequery DataFactory" );
					}
					setUserGivenSchema();
				}
				
			}
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);

		}
		if (this.forceConnectionRecycle) {
			try {
				if (logger.isDebugEnabled()) {
					logger.info("BDSPureQueryWriter.externalizeCheckpoint: forcing connection recycle, closing pureQuery Data interface and JDBC Connection");
				}
				if ( pqDataInterface != null ) pqDataInterface.close();
				jdbcConn.close();
			} catch (Throwable t) {
				throw new RuntimeException(t);
			}
		}

		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":externalizeCheckpointInformation");
		return "";
	}
	
	/**
	 * This method is invoked after a checkpoint has been taken. If forceConnectionRecycle
	 * is enabled then this method will open a new jdbcconnection and create a new pureQuery Data object
	 */
	public void intermediateCheckpoint() {	
		String methodName = "intermediateCheckpoint";
		if(logger.isDebugEnabled())
		   logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if (this.forceConnectionRecycle)
		{
			try {
				if (logger.isDebugEnabled()) {
					logger.info("BDSPureQueryWriter.intermediateCheckpoint: forcing connection recycle, opening JDBC Connection");
					logger.info("BDSPureQueryWriter.intermediateCheckpoint: ...and creating a new pureQuery Data object");
				}
				jdbcConn.open();
				// get Data implementation from pureQuery
				if ( dataInterfaceClass != null )
					pqDataInterface = (Data) ( DataFactory.getData( dataInterfaceClass, jdbcConn.getJDBCConnection() ) );
				else 
					pqDataInterface = (Data) ( DataFactory.getData( jdbcConn.getJDBCConnection() ) );
				if (logger.isDebugEnabled()) {
					logger.info( CLASSNAME + ": Data interface obtained from purequery DataFactory" );
				}
				setUserGivenSchema();
			} catch (Throwable t) {
				throw new RuntimeException(t);
			}
		}

		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":" + methodName);
		
	}

	/**
	 * Not implemented
	 */
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		// not sure we need to do anything here. I dont think there is any 
		// repositioning logic needed for insert/update logic.
	}
	/**
	 * Not implemented
	 */
	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {		
	}
	/**
	 * Not implemented
	 */
	public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {
		// not sure there is anything to do here.
	}
	/**
	 * Not implemented
	 */
	public void writeHeader(Object record) throws Exception {
		// not applicable
		
	}

}
