// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.sql.ResultSet;
import java.util.Properties;

import javax.naming.InitialContext;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;
import com.ibm.ws.batch.CursorHoldableJDBCReaderLocal;
import com.ibm.ws.batch.CursorHoldableJDBCReaderLocalHome;

/**
 * 
 * @author Neeraj
 *
 *  This class uses a Bean managed tx Stateful Session bean to load a cursor holdable non-XA 
 *  datasource (non-shareable). This pattern allows the resultset to live beyond the transaction
 *  commits thus improving performance.
 */



public class CursorHoldableBDSJDBCReader extends AbstractBatchDataStream 
	implements AbstractBatchDataInputStream {
	
	public static final String CLASSNAME = CursorHoldableBDSJDBCReader.class.getName();
	protected JDBCReaderPattern jdbcReaderImpl; 
	protected CursorHoldableJDBCReaderLocal sessionBeanJDBCReader;
	protected boolean prefetchedRecordHasBeenRead = false;
	protected Object record;
	protected ResultSet resultSet;
	protected String restartToken;
	protected int recordCount;
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;

	public static final String JAVA_PREFIX =  "java:comp/env/ejb/";
	
	/**
	 * Loads the CursorHoldable JDBC Reader bean and initializes it with the properties
	 * passed via the xJCL
	 */
	public void initialize(Properties props)
	{
		try 
		{
			logger = new BDSFWLogger(props);
			perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
			isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
			String appName = null; //props.getProperty(BatchConstants.DEFAULT_APPLICATION_NAME);
			//version prior to 6103 of batchruntime.jar do not provide the app name
			// as a property to the BDS hence it needs to be supplied via xJCL
			
			if(appName == null)
			{
				appName = props.getProperty(BDSFrameworkConstants.DEFAULT_APPLICATION_NAME_KEY);
			}

			String jndiName = JAVA_PREFIX  + appName +  "CursorHoldableJDBCReader";
			
			if(logger.isDebugEnabled())
				logger.debug("Loading ejb with JNDI: " + jndiName);
				
			Object hm =  new InitialContext().lookup(jndiName);
			
			if(logger.isDebugEnabled())
				logger.debug("Got CursorHoldableJDBCReaderHome " + hm);
				
			CursorHoldableJDBCReaderLocalHome sessionBeanJDBCReaderHome = (CursorHoldableJDBCReaderLocalHome)hm;
				 
			sessionBeanJDBCReader = sessionBeanJDBCReaderHome.create();
				
			sessionBeanJDBCReader.initialize(props);
				
			
			
			String jdbcReaderImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(jdbcReaderImplClass == null)
				jdbcReaderImplClass = props.getProperty("IMPLCLASS");		
			jdbcReaderImpl = (JDBCReaderPattern) loadClass(jdbcReaderImplClass);
			if(logger.isDebugEnabled())
				logger.info("Loaded JDBCReaderPattern "  + jdbcReaderImplClass);
			jdbcReaderImpl.initialize(props);
			
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
	}
	
	/**
	 * Delegates the positionAtInitialCheckpoint method to the sessionBeanJDBCReader
	 */
	

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {
		String methodName = "positionAtInitialCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try 
		{			
			recordCount = 0;
			String lookupQuery = jdbcReaderImpl.getInitialLookupQuery();
			sessionBeanJDBCReader.positionAtInitialCheckpoint(lookupQuery);
			
		
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
/**
 * Deleagates the positionAtCurrentCheckpoint method to the sessionBeanJDBCReader
 * after obtaining the restartQuery from the JDBCReaderPattern implementation class
 */
	public void positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {
		String methodName = "positionAtCurrentCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try 
		{
			String restartQuery = jdbcReaderImpl.getRestartQuery(restartToken);
			sessionBeanJDBCReader.positionAtCurrentCheckpoint(restartQuery);
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	/**
	 * fetches the next record from the session bean JDBC Reader and passes it
	 * to the JDBCReaderPattern implementation class with the fetchRecord method
	 */
	public Object read() throws Exception {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		recordCount++;
		ResultSet resultSet =  sessionBeanJDBCReader.getNext();
		Object rec = jdbcReaderImpl.fetchRecord(resultSet); 
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return rec;
	}
	
	/**
	 * Delegates the execution of the passed SQL query to the session bean JDBC Reader
	 * bean 
	 */
	public Object read(Object sql) throws Exception {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		Object result  =  sessionBeanJDBCReader.get(sql);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return result;
	}

	/**
	 * 
	 */
	public boolean hasNext() throws Exception {
		boolean hasNxt = false;
		String methodName = "hasNext";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		hasNxt = sessionBeanJDBCReader.hasNext();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return hasNxt;
	}
	
	/**
	 * Invokes open() on the session bean JDBC Reader
	 */
	public void open() throws BatchContainerDataStreamException {
		String methodName = "open";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		sessionBeanJDBCReader.open();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	/**
	 * Causes the session bean JDBC Reader to close the connection
	 */

	public void close() throws BatchContainerDataStreamException {

		sessionBeanJDBCReader.close();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.displayPerformanceStatisticsForEntity();
	}

	/**
	 * Invokes getRestartTokens() on the JDBCReaderPattern impl class
	 * just before a checkpoint is taken
	 */
	public String externalizeCheckpointInformation() {
		String methodName = "externalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try
		{			
			String retval = jdbcReaderImpl.getRestartTokens();	
			
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME + "." + methodName + "- token= |" + retval + "|");
			
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
			return retval ; 
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		
	}
	
	/**
	 * Obtains the restart query from the JDBCReaderPattern implementation class and 
	 * invokes the internalizeCheckpoint method on the session bean JDBC Reader
	 */
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + "." + methodName + "- token= |" + token + "|");
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try 
		{
			
			
			restartToken = token;
			String query = jdbcReaderImpl.getRestartQuery(token);
			if (logger.isDebugEnabled())
				logger.debug("CursolrHoldableJDBCReader.internalizeCheckpointInfo- query= |" + query + "|");
			
			sessionBeanJDBCReader.internalizeCheckpoint(query);

		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	/**
	 * Not implemented
	 */

	public Object fetchHeader() {
		// Not Applicable
		return null;
	}

}
