// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.sql.ResultSet;
import java.util.Properties;

/**
 * 
 * @author Snehal/Neeraj
 * 
 * This pattern is used to implement the business logic of processing data from
 *  JDBC. The administrative tasks of opening, closing the connection
 * is taken care of by the batch framework.
 * 
 * Prequisites:
 * For a Local JDBC database:
 * - Needs the following properties to be defined in the xJCL
 * jdbc_url, jdbc_driver, tablename and optionally user_id,pswd
 * E.g
 *  <bds>
 *		<logical-name>outputStream</logical-name>
 *		<props>
 *			<prop name="IMPLCLASS" value="com.ibm.websphere.batch.samples.creditsample.LocalDBReader"/>
 *          <prop name="jdbc_url" value="jdbc:derby:C:\\mysample\\CREDITREPORT"/>
 *          <prop name="jdbc_driver" value="org.apache.derby.jdbc.EmbeddedDriver"/>
 *          
 *          <prop name="user_id" value="myid"/>
 *          <prop name="pswd" value="mypwd"/>
 *          <prop name="debug" value="true"/>
 *		</props>
 *			<impl-class>com.ibm.websphere.batch.devframework.datastreams.patterns.LocalJDBCReader</impl-class>
 *		</bds>
 * 
 *  
 * For a Network JDBC database:
 * - Needs the following properties to be defined in the xJCL
 * ds_jndi_name, tablename
 * 
 * E.g
 * <bds>
 *			<logical-name>inputStream</logical-name>
 *			<props>
 *				<prop name="IMPLCLASS" value="com.ibm.websphere.batch.samples.creditsample.MYDBReader"/>
 *              <prop name="ds_jndi_name" value="jdbc/crreport"/>
 *              <prop name="debug" value="true"/>
 *			</props>
 *			<impl-class>com.ibm.websphere.batch.devframework.datastreams.patterns.JDBCReader</impl-class>
 * </bds>
 * 
 */

public interface JDBCReaderPattern {

	/**
	 * This method is invoked during the job setup phase. 
	 * 
	 * @param props properties provided in the xJCL 
	 */
	
	public void initialize(Properties props);
	
	/**
	 * This method should retrieve values for the various columns for the current row from the given resultset object.
	 * Typically this data would be used to populate an intermediate object which would be returned
	 * @param resultSet
	 * @return
	 */
	public Object fetchRecord(ResultSet resultSet);
	
	/**
	 * This method should return a SQL query that will be used during setup of the stream to retrieve all relevant
	 * data that would be processed part of the job steps
	 * @return object to be used during process step.
	 */
	public String getInitialLookupQuery();
	
	/**
	 * This method gets called during Job Restart. The restart token should be used to create an SQL query that will
	 * retrieve previously unprocessed records. 
	 * Typically the restart token would be the primary key in the table and the query would get all rows with 
	 * primary key value > restarttoken
	 * @param restartToken
	 * @return The restart query
	 */
	public String getRestartQuery(String restartToken);
	
	/**
	 * This method gets called just before a checkpoint is taken.
	 * @return The method should return a string value identifying the last record read by the stream.
	 */
	public String getRestartTokens();
		
}
