// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patterns;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.nio.channels.FileChannel;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.BDSFileByteWriter;

public class FileByteWriter extends BDSFileByteWriter {
	
	protected String fileName = "";	
	
	private boolean appendToFile = true;

	public static final String CLASSNAME = FileByteWriter.class.getName();
	
	
	/**
	 * If the AppendJobIdToFileName property is specified and set to true then
	 * appends the jobid to the filename
	 */
	protected void initialize(Properties props) {
		this.fileName = getRequiredProperty(props,BDSFrameworkConstants.FILENAME_KEY);
		
		// If isJobRestart property is set by the runtime and is false do not append to file
		// otherwise always append (property not specified or property specified and set to true
		String jobRestartedStr = (String)props.getProperty(BDSFrameworkConstants.IS_JOB_RESTART);
		if(jobRestartedStr != null && jobRestartedStr.equals("false")) {
			
			isRestart = false;
		} else {
			isRestart = true;
			appendToFile = true;
		}
		
		String appendToFileStr = (String)props.getProperty(BDSFrameworkConstants.APPEND);
		if(appendToFileStr != null && appendToFileStr.equals("false")&& !isRestart) {
			
			appendToFile = false;
		}
		
		String appendJobIdProp = (String)props.get(BDSFrameworkConstants.APPEND_JOBID_TO_FILE_NAME_KEY);
		
		
		if (appendJobIdProp == null || appendJobIdProp.equalsIgnoreCase("true"))
		{						
			String jobStepId = props.get(BDSFrameworkConstants.JOB_STEP_ID).toString();
			
			String jobId = jobStepId.substring(0, jobStepId.indexOf('/'));
			
			jobId = jobId.replace('/', '.');
			jobId = jobId.replace(':', '.');
			jobId = jobId.replace('$', '.');
			fileName += "." + jobId;			
		}
		
		super.initialize(props);
	}

	/**
	 * Uses the FileChannel to obtain the current file pointer position
	 */
	protected long getCurrentPosition() {
		String methodName = "getCurrentPosition";
		if(logger.isDebugEnabled())
			logger.debug("In " + CLASSNAME + ":" + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		long position = 0;
		try
		{
			writer.flush();
			if(fos.getChannel().isOpen())
			{	
				
				position =  fos.getChannel().position();
			}
			else
			{
				position = currentPosition;
			}
			
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return position;
	}

	/**
	 * Opens the file
	 */
	public void open() throws BatchContainerDataStreamException {
		String methodName = "open";
		boolean writeHeader = false;
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try
		{
			File outputFile = new File(fileName);
			if (!outputFile.exists())
			{	
				outputFile.createNewFile();
				writeHeader = true;
				fos = new FileOutputStream(outputFile);			
				writer = new BufferedOutputStream(fos);
				if(writeHeader)
					byteWriterPattern.writeHeader(writer);
			}
			else
			{
				if(appendToFile)
				{	
					fos = new FileOutputStream(outputFile,true);			
					writer = new BufferedOutputStream(fos);
					if(writeHeader)
						byteWriterPattern.writeHeader(writer);
					
				}
				else
				{
					fos = new FileOutputStream(outputFile);			
					writer = new BufferedOutputStream(fos);
					if(writeHeader)
						byteWriterPattern.writeHeader(writer);
					
				}
			}
			
			
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
			if(logger.isDebugEnabled())
				logger.debug(CLASSNAME + ":open Opened file: " + fileName);
				
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}

	}
	
	/**
	 * Closes the file
	 */
	public void close() throws BatchContainerDataStreamException {
		
		try
		{
			FileChannel streamChannel = fos.getChannel();			
			streamChannel.truncate(getCurrentPosition());	
			fos.getChannel().force(true);
			writer.flush();
			super.close();
			if(logger.isDebugEnabled())
				logger.debug(CLASSNAME + ":close Closed file");
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}
	}

	/**
	 * Truncates the file based on the position passed. Used during a job restart
	 * this method will allow the new writes to occur exactly after the last
	 * checkpoint was taken 
	 */
	protected void setPosition(long position) {
		String methodName = "setPosition";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try
		{			
			fos.getChannel().truncate(position);
			fos.getChannel().force(true);
			writer.flush();
			if(logger.isDebugEnabled())
				logger.debug( CLASSNAME + ":" + methodName + "->" + position);
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

}
