// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patterns;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.nio.channels.FileChannel;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.BDSFileWriter;
 
public class TextFileWriter extends BDSFileWriter {

	private FileOutputStream fos;
	protected String fileName = "";
	protected boolean appendToFile = true;
	public static final String CLASSNAME = TextFileWriter.class.getName();
	
	/**
	 * Closes the file
	 */
	public void close() throws BatchContainerDataStreamException {
		try
		{									
			currentPosition = getCurrentPosition();
			FileChannel streamChannel = fos.getChannel();		
			streamChannel.truncate(currentPosition);	
			writer.flush();
			writer.close();
			if(perfAnalyzer.isEnabled())
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}
	} 

	/**
	 * Sets the FileChannel position based on the value passed
	 *  and truncates the file. This allows writing to the file to resume
	 *  from the last checkpoint
	 */
	public void setPosition(long position)
	{
		try {
			//No op always append at bottom of file
			logger.debug("Entering TextFileWriter:setPosition" + position);
			fos.getChannel().truncate(position);
			fos.getChannel().force(true);
			writer.flush();
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
	}

	/**
	 * Gets the current file pointer position
	 */
	
	public long getCurrentPosition()
	{	
		String methodName = "getCurrentPosition";
		long position = 0;
		if(perfAnalyzer.isEnabled())
			perfAnalyzer.startMeasurement(methodName);
		try {
			if(fos.getChannel().isOpen())
			{	
				writer.flush();
				position = fos.getChannel().position();
			}
			else
			{
				position = currentPosition;
			}
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}	
		if(perfAnalyzer.isEnabled())
			perfAnalyzer.endMeasurementAndSave(methodName);
		return position;
		
	}

	/**
	 * If this is a restart and AppendJobIdToFileName property is set to false then
	 * the jobid is not appended to the file. By default job id is appended to the filename
	 */
	public void initialize(Properties props) {
		
		logger = new BDSFWLogger(props);

		this.fileName = getRequiredProperty(props,BDSFrameworkConstants.FILENAME_KEY);
		// If isJobRestart property is set by the runtime and is false do not append to file
		// otherwise always append (property not specified or property specified and set to true
	
		String jobRestartedStr = (String)props.getProperty(BDSFrameworkConstants.IS_JOB_RESTART);
		if(jobRestartedStr != null && jobRestartedStr.equals("false")) {
			
			isRestart = false;
		} else {
			isRestart = true;
			appendToFile = true;
		}
		
		String appendToFileStr = (String)props.getProperty(BDSFrameworkConstants.APPEND);
		if(appendToFileStr != null && appendToFileStr.equals("false") && !isRestart) {
			
			appendToFile = false;
		}
		
		String appendJobIdProp = (String)props.get(BDSFrameworkConstants.APPEND_JOBID_TO_FILE_NAME_KEY);
		if (appendJobIdProp == null || appendJobIdProp.equalsIgnoreCase("true"))
		{						
			String jobStepId = props.get(BDSFrameworkConstants.JOB_STEP_ID).toString();
			
			String jobId = jobStepId.substring(0, jobStepId.indexOf('/'));
			
			jobId = jobId.replace('/', '.');
			jobId = jobId.replace(':', '.');
			jobId = jobId.replace('$', '.');
			fileName += "." + jobId;				
			
		}
		super.initialize(props);
				
	}

	/**
	 * If the file does not exist create a new file
	 * if it exists and this is a restart then open the file in append mode
	 * else overwrite existing file
	 */
	public void open() throws BatchContainerDataStreamException {
		boolean writeHeader = false;
		boolean isRestart = false;
		String methodName = "open";
		if(perfAnalyzer.isEnabled())
			perfAnalyzer.startMeasurement(methodName);
		try
		{
			File outputFile = new File(fileName);
			if (!outputFile.exists())
			{	
				outputFile.createNewFile();
				writeHeader = true;
				
				fos = new FileOutputStream(fileName);
				if(logger.isDebugEnabled())
					logger.debug("created new file" + fileName);
				
			} else	{
				
				
				if(appendToFile) {	
					fos = new FileOutputStream(fileName,true);
					//isRestart = true;
					if(logger.isDebugEnabled())
						logger.debug("Opened file in append mode");
				} else	{
					fos = new FileOutputStream(fileName);
					//isRestart = false;
					if(logger.isDebugEnabled())
						logger.debug("Opened file in non-append mode");
					
				}
			}
			
			
		
						
			if (this.fileEncoding != null)
			{
				writer = new BufferedWriter(new OutputStreamWriter(fos, this.fileEncoding));
				
			}
			else
			{	
				writer = new BufferedWriter(new OutputStreamWriter(fos));
			}			
			if(writeHeader)
				fileWriterImpl.writeHeader(writer);
			
			//if(isRestart)
				//writer.newLine();
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(perfAnalyzer.isEnabled())
			perfAnalyzer.endMeasurementAndSave(methodName);

	}

	
}
