// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patterns;

import java.util.Properties;

import com.ibm.jzos.ZFile;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.RecordOrientedDatasetReaderPattern;

public class ZFileRecordOrientedDataReader extends AbstractBatchDataStream 
	implements AbstractBatchDataInputStream {

	public static final String CLASSNAME = ZFileRecordOrientedDataReader.class.getName();
	protected String dsname = "";
	protected String ds_parameters = "rb,recfm=fb,type=record,lrecl=80";
	protected ZFile zFile;
	protected RecordOrientedDatasetReaderPattern recordOrientedReader; 
	protected boolean prefetchedRecordHasBeenRead = false;
	protected Object record;
	 
	protected Properties props;		
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer;
	protected boolean isPerformanceMeasurementEnabled = false;
	private boolean useSeekAndTell = true;
	
	protected void initialize(Properties props) {
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		
		this.props = props;		
		dsname = getRequiredProperty(props,BDSFrameworkConstants.DSNAME_KEY);
		//if (prop != null)
		//{
		//	this.dsname = prop;
		//}
		
		String prop = props.getProperty(BDSFrameworkConstants.DSParameters_KEY);
		if (prop != null)
		{
			this.ds_parameters = prop;
		}
		
		String largeFileSupportStr = props.getProperty(BDSFrameworkConstants.LARGE_DATASET_SUPPORT_KEY);
		if(largeFileSupportStr != null && largeFileSupportStr.equalsIgnoreCase("true")) {
			useSeekAndTell   = false;
		}
		int isNoSeek = ds_parameters.indexOf("noseek");
		if(isNoSeek != -1) {
			useSeekAndTell = false;
		}
		
				
		String implClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);	
		// For backward compatibility
		if(implClass == null)
			implClass = getRequiredProperty(props,"IMPLCLASS");	
		recordOrientedReader = (RecordOrientedDatasetReaderPattern) loadClass(implClass);
		recordOrientedReader.initialize(props);	
	}
	public boolean hasNext() throws Exception {	
		String methodName = "hasNext";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		boolean retval = false;
		// Did I prefetch a record in the previous read?

		// I already have a value set to the record object.  
		if (record != null)
		{
			// Have I already processed this value?
			if (prefetchedRecordHasBeenRead)
			{	
				// yes I have. go get another one.
				record = read();
				
				// was there any data left?
				if (record == null)
				{
					// no, there was no data left.
					retval = false;
				}
				else
				{
					// yes, there was data left. We've therefore prefetched the data.
					prefetchedRecordHasBeenRead = false;
					retval = true;
				}
			}
		}
		else
		{
			// I don't have data prefetched. This could be the first read we hit. Go get a record.
			record = read();
			
			// was there any data left?
			if (record == null)
			{
				// no, there wasn't any data left. 
				retval = false;
			}
			else
			{
				// yes, there was data left. We've therefore just prefetched it. 
				retval = true;
				prefetchedRecordHasBeenRead = false;
			}
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + "." + methodName + "=" + retval);
		return retval;
	}

	public Object read() throws Exception {	
		Object returnValue = null;
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if (!prefetchedRecordHasBeenRead) {
			if (record != null) {
				prefetchedRecordHasBeenRead = true;
				returnValue = record;
			} else {
				returnValue = recordOrientedReader.fetchRecord(zFile);
			}	
		} else {
			returnValue = recordOrientedReader.fetchRecord(zFile);		
		}	
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return returnValue;
	}
	
	
	public Object read(Object parameters) throws Exception {
		throw new Exception("Unsupported");
		
	}

	public void close() throws BatchContainerDataStreamException {

		try {
			zFile.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
	}

	public String externalizeCheckpointInformation() {
		String methodName = "externalizeCheckpointInformation";
		String returnValue = null;
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try {
			
			String retval = String.valueOf(this.getCurrentPosition());
			
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME +"." + methodName + "- token= |" + retval + "|");
			
			returnValue = String.valueOf(this.getCurrentPosition());		
		} catch (Throwable t) {
			throw new RuntimeException(t);

		}		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return returnValue;

	}


	public void internalizeCheckpointInformation(String token)
			throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		try {
			
			if (logger.isDebugEnabled())
				logger.debug("ZFileRecordOrientedDataReader.externalizeCheckpointInfo - token= |" + token + "|");
			
			if (token == null)
				this.positionAtInitialCheckpoint();
			else
				this.setPosition(Long.valueOf(token).longValue());
			
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);

	}

	public void open() throws BatchContainerDataStreamException {
		String methodName = "open";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try
		{						
			zFile = new ZFile("//'" + dsname + "'", ds_parameters);     						 		
			this.recordOrientedReader.processHeader(zFile);
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}     
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	public void positionAtInitialCheckpoint()
			throws BatchContainerDataStreamException {				
	}

	public void positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {		
	}


	public long getCurrentPosition() {
		try
		{
			if(this.useSeekAndTell)
				return zFile.tell();
			else {
				return zFile.getRecordCount();
			}
		}
		catch (Throwable t)
		{
			throw new RuntimeException("Failed to retrieve the current position of the zFile.", t);
		}
	}
	
	public void setPosition(long position) {
		try {
			if(this.useSeekAndTell) {
				zFile.seek(position,0);
			} else {
				for(int i=0; i< position; i++) {
					read();
				}
			}
				
		} 
		catch (Throwable t) {			
			throw new RuntimeException("Failed to set position of zFile", t);
		}
	}

	public Object fetchHeader() {
		Object header = null;
		String methodName = "fetchHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		header = recordOrientedReader.fetchHeader();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return header;
	}
}
