// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.datastreams.patterns;

import java.util.Properties;

import com.ibm.jzos.ZFile;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataStream;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.RecordOrientedDatasetWriterPattern;

public class ZFileRecordOrientedDataWriter extends AbstractBatchDataStream 
	implements AbstractBatchDataOutputStream {

	public static final String CLASSNAME = ZFileRecordOrientedDataWriter.class.getName();
	protected String dsname = "";
	//protected String ds_parameters = "wb,recfm=fb,type=record,lrecl=80";
	protected String ds_parameters = "ab,recfm=fb,type=record,lrecl=80";
	protected String ds_parameters_restart = "rb+,recfm=fb,type=record,lrecl=80";
	protected boolean headerWritten = false;
	protected ZFile zFile;
	protected RecordOrientedDatasetWriterPattern recordOrientedWriter;
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer;
	protected boolean isPerformanceMeasurementEnabled = false;
	protected boolean isRestart = false;
	private boolean useSeekAndTell = true;
	
	

	protected void initialize(Properties props) {
		
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		
		dsname = getRequiredProperty(props,BDSFrameworkConstants.DSNAME_KEY);
		/*if (prop != null)
		{
			this.dsname = prop;
		}	*/			
		
		String prop = props.getProperty(BDSFrameworkConstants.DSParameters_KEY);
		if (prop != null)
		{
			this.ds_parameters = prop;
		}	
		
		prop = props.getProperty(BDSFrameworkConstants.DSParametersRestart_KEY);
		if (prop != null)
		{
			this.ds_parameters_restart = prop;
		}	
		
		String jobRestartedStr = (String)props.getProperty(BDSFrameworkConstants.IS_JOB_RESTART);
		if(jobRestartedStr != null && jobRestartedStr.equals("true")) {
			
			isRestart = true;
		}
		
		String largeFileSupportStr = props.getProperty(BDSFrameworkConstants.LARGE_DATASET_SUPPORT_KEY);
		if(largeFileSupportStr != null && largeFileSupportStr.equalsIgnoreCase("true")) {
			useSeekAndTell  = false;
		}
		int isNoSeek = ds_parameters.indexOf("noseek");
		if(isNoSeek != -1) {
			useSeekAndTell = false;
		}
		
		String implClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
		// For backward compatibility
		if(implClass == null)
			implClass = getRequiredProperty(props,"IMPLCLASS");	
		recordOrientedWriter = (RecordOrientedDatasetWriterPattern) loadClass(implClass);
		recordOrientedWriter.initialize(props);	
	}
	public void flush() throws Exception {
		zFile.flush();		
	}

	public void write(Object record) throws Exception {
		String methodName = "write";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		if (!headerWritten)
		{
			long currPos = this.getCurrentPosition();
			if (currPos == 0 ) 
			{
				// This is our first write to the file. Need to create some header information.
				recordOrientedWriter.writeHeader(zFile);
				headerWritten = true;
			} else if(currPos > 0){
				// do not check for header any more as no header has been written
				headerWritten = true;
			}
		}
		recordOrientedWriter.writeRecord(zFile, record);
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		
	}

	public void close() throws BatchContainerDataStreamException {
		
		try 
		{
			zFile.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		}
		catch (Throwable t)
		{
			throw new BatchContainerDataStreamException(t);
		}
	}

	public String externalizeCheckpointInformation() {	
		
		String retval = String.valueOf(getCurrentPosition());
		
		if (logger.isDebugEnabled())
			logger.debug("ZFileRecordOrientedDataWriter.externalizeCheckpointInfo- token= |" + retval + "|");
		
		return String.valueOf(getCurrentPosition());
	}
	public void internalizeCheckpointInformation(String token) throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + "." + methodName + "- token= |" + token + "|");
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		isRestart = true;
		try
		{
			if (token == null)
				positionAtInitialCheckpoint();
			else
				setPosition(Long.valueOf(token).longValue());
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	public void open() throws BatchContainerDataStreamException {
		String methodName = "open";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try {        
			if(!isRestart) {
				zFile = new ZFile("//'" + dsname + "'", ds_parameters);
			} else {
				zFile = new ZFile("//'" + dsname + "'", ds_parameters_restart);
			}
        } catch (Throwable t) {
            throw new BatchContainerDataStreamException(t);
        }				
        if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	public long getCurrentPosition() {
		try
		{
			flush();
			if(useSeekAndTell)
				return zFile.tell();
			else {
				return zFile.getRecordCount();
			}
		}
		catch (Throwable t)
		{
			throw new RuntimeException("Failed to retrieve the current position of the zFile.", t);
		}
	}
	
	public void setPosition(long position) {
		try {
			if(useSeekAndTell) {
				zFile.seek(position,0);
			} else {
				for(int i=0; i<position; i++) {
					byte[] buf = new byte[zFile.getLrecl()];			
					zFile.read(buf);
				}	
				flush();
				
			}
		} 
		catch (Throwable t) {			
			throw new RuntimeException("Failed to set position of zFile", t);
		}
	}

	public void positionAtInitialCheckpoint() throws BatchContainerDataStreamException {		
	}

	public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {		
	}

	public void writeHeader(Object header) throws Exception {
		String methodName = "writeHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if(!isRestart)
			recordOrientedWriter.writeHeader(zFile, header);
		
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
}
