// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.steps.technologyadapters;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;



import com.ibm.websphere.batch.BatchConstants;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.BatchDataStreamMgr;
import com.ibm.websphere.batch.BatchJobStepInterface;
import com.ibm.websphere.batch.JobStepID;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;

/**
 * @author Neeraj
 * This is a simple batch step that takes exactly 1 inputstream and 1 outputstream
 * In each iteration of the step it reads a record from the inputstream
 * invokes the BatchRecordProcessor.processRecord and then calls the outputStream
 * with the returned value.
 * 
 */
public class GenericParallelXDBatchStep implements BatchJobStepInterface {

	public static final int RC_OK = 0;
	public static final int RC_WARNING = 4;
	public static final int RC_ERROR = 8;
	public static final int RC_FATAL = 12;
	private static Object outputlock = new Object();
	private static Object counterlock = new Object();
	public static final String CLASSNAME = GenericXDBatchStep.class.getName();
	
    protected Properties props;
    protected AbstractBatchDataInputStream inputStream;
	protected AbstractBatchDataOutputStream outputStream;
	protected BatchRecordProcessor batchRecordProcessor;
	
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer;
	protected boolean isPerformanceMeasurementEnabled = false;
	
	protected ThreadLocal batchRecordProcessorHandle;	
	protected long time;
	protected Object _header;
	protected boolean _headerWritten;
	protected int _noOfParallelProcessors = 3;
	protected int _activeTaskCount = 3;
	ExecutorService _cachedThreadPool = Executors.newCachedThreadPool();
    protected ConcurrentLinkedQueue outputQueue = new ConcurrentLinkedQueue();
    protected void setReturnCode(int returnCode) {
    	
    }

    /**
     * Convenience method to return a property value.
     * @param key
     *   Property key.
     * @return
     *   Property value.
     */
    protected String getProperty(String key) {
        return getProperties().getProperty(key);
    }

    /**
     * Convenience method to return a property value.
     * @param key
     *   Property key.
     * @param defaultValue
     *   Default value if property was not set.
     * @return
     *   Property value.
     */
    protected String getProperty(String key, String defaultValue) {
        String value = getProperty(key);
        if (value == null) {
            value = defaultValue;
        }
        return value;
    }

    /**
     * Convenience method to return a property value.
     * Throws an exception if the property was not set, or was
     * set to an empty value.
     * 
     * @param key
     *   Property key.
     * @return
     *   Property value.
     *  
     * @throws RuntimeException
     *   No value was specified for the given key.
     */
    protected String getRequiredProperty(String key) {
        String value = getProperties().getProperty(key);
        if (value == null) {
            throw new RuntimeException("Missing required property '" + key + "'");
        }
        value = value.trim();
        if (value.length() == 0) {
            throw new RuntimeException("Required property '" + key + "' has empty value");
        }
        return value;
    }

    public void createJobStep() {
    	String methodName = "createJobStep";
        try {            	        	
        	
        	logger = new BDSFWLogger(this.getProperties());
        	perfAnalyzer = new PerformanceAnalyzer(CLASSNAME + "." + "processRecord", props);
        	isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
        	
        	if (logger.isDebugEnabled())
        		logger.debug("entering createJobStep()- jobId="+getJobID());

        	time = System.currentTimeMillis();
            
            inputStream = (AbstractBatchDataInputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.INPUT_STREAM_KEY, getStepID());
            
    		outputStream = (AbstractBatchDataOutputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.OUTPUT_STREAM_KEY, getStepID());
    		
    		if (inputStream == null)
    			logger.error("No Inputstream defined for the batch step!");
    		
    		if (outputStream == null)
    			logger.error("No Outputstream defined for the batch step!");
    		
    		try {
                
            	_header = inputStream.fetchHeader();
            	if (logger.isDebugEnabled())
            		logger.debug("GenericXDBatchStep: Got header" + _header);
            
            } catch (Exception e) {
            	// No header data
            	_header = null;
            	
            }
    		
    		String batchRecordProcessorImplClass = getRequiredProperty(BDSFrameworkConstants.BATCH_RECORD_PROCESSOR_KEY);
    		batchRecordProcessor = (BatchRecordProcessor)loadClass(batchRecordProcessorImplClass);
    		batchRecordProcessor.initialize(props);    		    	
    		
    		String noOfParallelProcessorsStr = getRequiredProperty("PARALLEL_PROCESSOR_COUNT");
    		_noOfParallelProcessors = Integer.parseInt(noOfParallelProcessorsStr);
    		
    		batchRecordProcessorHandle = new ThreadLocal();
    		batchRecordProcessorHandle.set(batchRecordProcessor);    		
    		
        	if (logger.isDebugEnabled())
        		logger.debug("exiting createJobStep");

        }
        catch(Exception e) {
            throw new RuntimeException("Unexpected error in step setup", e);
        }
    }

    /**
     * Perform cleanup and pass return code back to LREE.
     */
    public int destroyJobStep() { 
    	
    	if (logger.isDebugEnabled())
    		logger.debug("entering destroyJobStep");
    	int returnCode = 0;
    	batchRecordProcessor = (BatchRecordProcessor) batchRecordProcessorHandle.get();
    	
    	returnCode = batchRecordProcessor.completeProcessing();
    	if(isPerformanceMeasurementEnabled)
			perfAnalyzer.displayPerformanceStatisticsForEntity();
                
        logger.info("GenericXDBatchStep.destroyStep()- Total Execution Time: " + (System.currentTimeMillis() - time));
        
    	if (logger.isDebugEnabled())
    		logger.debug( "destroyJobStep");

        return returnCode;
    }

    public Properties getProperties() {
        return props;
    }

    public int processJobStep() {    	    	
    	String methodName = "processJobStep";
    	List recordList = new ArrayList();
    	int returnValue = -1;
    	
    	if (logger.isDebugEnabled())
    		logger.debug("processJobStep");

    	try
    	{    		
    		if (inputStream == null)
    			logger.error("No Inputstream defined for the batch step!");
    		
    		if (outputStream == null)
    			logger.error("No Outputstream defined for the batch step!");
    		
    		if(_header != null && !_headerWritten ) {
    			if (logger.isDebugEnabled()) 
    				logger.debug("GenericXDBatchStep:processJobStep Writing header" + _header);
    			outputStream.writeHeader(_header);
    			_headerWritten = true;
    		}
    		
            if (inputStream.hasNext()) {
            	
            	if (logger.isDebugEnabled())
            		logger.debug("Inputstream has data to process.");
            	
            	for(int i=0; i<this._noOfParallelProcessors; i++) {
            		Object rec = inputStream.read();
            		 if (logger.isDebugEnabled())
                 		logger.debug("read record = " + rec);
            		if(rec != null) {
            			recordList.add(rec);
            		} else {
            			break;
            		}
            	}
            	_activeTaskCount = _noOfParallelProcessors;
            	_parallelProcessRecords(recordList);
            	do {
            		Thread.sleep(100);
            	} while(_activeTaskCount > 0);	
            	
            	recordList.clear();
            	if (logger.isDebugEnabled())
            		logger.debug("finish parallel processing");

               
            	
               
                returnValue = BatchConstants.STEP_CONTINUE;
            }
            else {
            	
            	if (logger.isDebugEnabled())
            		logger.debug("No more records to process in Input stream. Completing step.");

                returnValue = BatchConstants.STEP_COMPLETE;
            }
        }
    	catch (BatchContainerDataStreamException e)
    	{            	
    		throw new RuntimeException("failed to obtain batch input/output streams", e);
    	}        
        catch(Exception e) {
            throw new RuntimeException("Unexpected error in batch loop", e);
        }
        return returnValue;
    }

    private void _parallelProcessRecords(List recordList) {
    	if (logger.isDebugEnabled())
    		logger.debug("Entering _parallelProcessRecords");

    	for(int i=0; i< recordList.size(); i++) {
    		
    		RecordProcessor recordProcessor = new RecordProcessor(recordList.get(i));
    		_cachedThreadPool.execute(recordProcessor);
    	}
    	
    	if (logger.isDebugEnabled())
    		logger.debug("Exiting _parallelProcessRecords");
    	
    	
		
	}

	protected Object processRecord(Object record) throws Exception
    {
    	String methodName = "processRecord";
    	if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
    	Object rec = null;
    	rec = batchRecordProcessor.processRecord(record);
    	if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
    	return rec;
    }

    /**
     * Called by LREE to set our properties as specified in xJCL.
     */
    public void setProperties(Properties props) {
        this.props = props;
        props.put("JobStepId", getJobID());
    }

   
    public String getJobID() {
        return  (String) getProperties().get(BatchConstants.JOB_ID); //JobStepContextMgr.getContext().getJobID();// return Integer.toString(hashCode());

    }
    

    public String getStepID() {
    	String jobID = (String) getProperties().get(BatchConstants.JOB_ID);
        String stepID = (String) getProperties().get(BatchConstants.STEP_ID);

        //System.out.println("SSA: GenericXDBatchStep.getStepID(): jobID= " + jobID + ", stepID= " + stepID);
        // System.out.println("getStepID jobID " + jobID + " stepID " + stepID);
    	// the bds logical name and the jobid/stepid are needed to retrieve the bds. 
        JobStepID jsId = new JobStepID( jobID, stepID );
        
        return  jsId.getJobstepid(); //JobStepContextMgr.getContext().getJobID();// return Integer.toString(hashCode());

    }
    
    protected Object loadClass(String className)
	{
		try
		{
			Object retval = Thread.currentThread().getContextClassLoader().loadClass(className).newInstance();
			return retval;
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}		
	}
    
    class RecordProcessor implements Runnable {

    	Object _record;
    	public RecordProcessor(Object record) {
    		_record = record;
    	}
		public void run() {
			
			try {
				_record = processRecord(_record);
				if (logger.isDebugEnabled())
		     		logger.debug("processed record = " + _record);
		         
		         if (_record != null) {
		         	     
		        //	 synchronized(outputlock) {
		        	 	outputStream.write(_record);
		        	 	if (logger.isDebugEnabled())
		            		logger.debug("wrote output " + _record);
		        //	 }
		        	 synchronized(counterlock) {
		        		 _activeTaskCount--;
		        	 }	 
		         		
		         } 
			} 
			catch (BatchContainerDataStreamException e)
	    	{            	
	    		throw new RuntimeException("failed to obtain batch input/output streams", e);
	    	}        
	        catch(Exception e) {
	            throw new RuntimeException("Unexpected error in batch loop", e);
	        }
	         
	     	
			
		}
		protected Object processRecord(Object record) throws Exception
	    {
			
	    	String methodName = "processRecord";
	    	if(isPerformanceMeasurementEnabled)
				perfAnalyzer.startMeasurement(methodName);
	    	Object rec = null;
	    	
	    	rec = batchRecordProcessor.processRecord(record);
	    	if(isPerformanceMeasurementEnabled)
				perfAnalyzer.endMeasurementAndSave(methodName);
	    	return rec;
	    }

    	
    }
}
