// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.steps.technologyadapters;

import java.util.Properties;

import com.ibm.websphere.batch.BatchConstants;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.BatchDataStreamMgr;
import com.ibm.websphere.batch.BatchJobStepInterface;
import com.ibm.websphere.batch.JobStepID;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;

/**
 * @author Antani/Neeraj
 * This is a simple batch step that takes exactly 1 inputstream and 1 outputstream
 * In each iteration of the step it reads a record from the inputstream
 * invokes the BatchRecordProcessor.processRecord and then calls the outputStream
 * with the returned value.
 * 
 */
public class GenericXDBatchStep implements BatchJobStepInterface {

	public static final int RC_OK = 0;
	public static final int RC_WARNING = 4;
	public static final int RC_ERROR = 8;
	public static final int RC_FATAL = 12;
	
	public static final String CLASSNAME = GenericXDBatchStep.class.getName();
	
    protected Properties props;
    protected AbstractBatchDataInputStream inputStream;
	protected AbstractBatchDataOutputStream outputStream;
	protected BatchRecordProcessor batchRecordProcessor;
	
	protected BDSFWLogger logger;
	protected PerformanceAnalyzer perfAnalyzer;
	protected boolean isPerformanceMeasurementEnabled = false;
	
	protected ThreadLocal batchRecordProcessorHandle;	
	protected long time;
	protected Object _header;
	protected boolean _headerWritten;
    
    protected void setReturnCode(int returnCode) {
    	
    }

    /**
     * Convenience method to return a property value.
     * @param key
     *   Property key.
     * @return
     *   Property value.
     */
    protected String getProperty(String key) {
        return getProperties().getProperty(key);
    }

    /**
     * Convenience method to return a property value.
     * @param key
     *   Property key.
     * @param defaultValue
     *   Default value if property was not set.
     * @return
     *   Property value.
     */
    protected String getProperty(String key, String defaultValue) {
        String value = getProperty(key);
        if (value == null) {
            value = defaultValue;
        }
        return value;
    }

    /**
     * Convenience method to return a property value.
     * Throws an exception if the property was not set, or was
     * set to an empty value.
     * 
     * @param key
     *   Property key.
     * @return
     *   Property value.
     *  
     * @throws RuntimeException
     *   No value was specified for the given key.
     */
    protected String getRequiredProperty(String key) {
        String value = getProperties().getProperty(key);
        if (value == null) {
            throw new RuntimeException("Missing required property '" + key + "'");
        }
        value = value.trim();
        if (value.length() == 0) {
            throw new RuntimeException("Required property '" + key + "' has empty value");
        }
        return value;
    }

    public void createJobStep() {
    	String methodName = "createJobStep";
        try {            	        	
        	
        	logger = new BDSFWLogger(this.getProperties());
        	perfAnalyzer = new PerformanceAnalyzer(CLASSNAME + "." + "processRecord", props);
        	isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
        	
        	if (logger.isDebugEnabled())
        		logger.debug("entering createJobStep()- jobId="+getJobID());

        	time = System.currentTimeMillis();
            
            inputStream = (AbstractBatchDataInputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.INPUT_STREAM_KEY, getStepID());
            
    		outputStream = (AbstractBatchDataOutputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.OUTPUT_STREAM_KEY, getStepID());
    		
    		if (inputStream == null)
    			logger.error("No Inputstream defined for the batch step!");
    		
    		if (outputStream == null)
    			logger.error("No Outputstream defined for the batch step!");
    		
    		try {
                
            	_header = inputStream.fetchHeader();
            	if (logger.isDebugEnabled())
            		logger.debug("GenericXDBatchStep: Got header" + _header);
            
            } catch (Exception e) {
            	// No header data
            	_header = null;
            	
            }
    		
    		String batchRecordProcessorImplClass = getRequiredProperty(BDSFrameworkConstants.BATCH_RECORD_PROCESSOR_KEY);
    		batchRecordProcessor = (BatchRecordProcessor)loadClass(batchRecordProcessorImplClass);
    		batchRecordProcessor.initialize(props);    		    	
    		
    		batchRecordProcessorHandle = new ThreadLocal();
    		batchRecordProcessorHandle.set(batchRecordProcessor);    		
    		
        	if (logger.isDebugEnabled())
        		logger.debug("exiting createJobStep");

        }
        catch(Exception e) {
            throw new RuntimeException("Unexpected error in step setup", e);
        }
    }

    /**
     * Perform cleanup and pass return code back to LREE.
     */
    public int destroyJobStep() { 
    	
    	if (logger.isDebugEnabled())
    		logger.debug("entering destroyJobStep");
    	int returnCode = 0;
    	if(batchRecordProcessorHandle != null) {
    	batchRecordProcessor = (BatchRecordProcessor) batchRecordProcessorHandle.get();
	    	if(batchRecordProcessor == null) {
	    		String batchRecordProcessorImplClass = getRequiredProperty(BDSFrameworkConstants.BATCH_RECORD_PROCESSOR_KEY);
	    		batchRecordProcessor = (BatchRecordProcessor)loadClass(batchRecordProcessorImplClass);
	    		batchRecordProcessor.initialize(props);    		    	
	    		returnCode = batchRecordProcessor.completeProcessing();	
	    	} else {
	    		returnCode = batchRecordProcessor.completeProcessing();	
	    	}
	    	
	    	
    	}
    	
    	if(isPerformanceMeasurementEnabled)
			perfAnalyzer.displayPerformanceStatisticsForEntity();
                
        logger.info("GenericXDBatchStep.destroyStep()- Total Execution Time: " + (System.currentTimeMillis() - time));
        
    	if (logger.isDebugEnabled())
    		logger.debug( "destroyJobStep");

        return returnCode;
    }

    public Properties getProperties() {
        return props;
    }

    public int processJobStep() {    	    	
    	String methodName = "processJobStep";
    	int returnValue = -1;
    	
    	if (logger.isDebugEnabled())
    		logger.debug("processJobStep");

    	try
    	{    		
    		if (inputStream == null)
    			logger.error("No Inputstream defined for the batch step!");
    		
    		if (outputStream == null)
    			logger.error("No Outputstream defined for the batch step!");
    		
    		if(_header != null && !_headerWritten ) {
    			if (logger.isDebugEnabled()) 
    				logger.debug("GenericXDBatchStep:processJobStep Writing header" + _header);
    			outputStream.writeHeader(_header);
    			_headerWritten = true;
    		}
    		
            if (inputStream.hasNext()) {
            	
            	if (logger.isDebugEnabled())
            		logger.debug("Inputstream has data to process.");

                Object record = inputStream.read(); 
            	
                if (logger.isDebugEnabled())
            		logger.debug("read record = " + record);
            	
                record = processRecord(record);
                
            	if (logger.isDebugEnabled())
            		logger.debug("processed record = " + record);
                
                if (record != null) {
                	              	
                		outputStream.write(record);
                } 
                returnValue = BatchConstants.STEP_CONTINUE;
            }
            else {
            	
            	if (logger.isDebugEnabled())
            		logger.debug("No more records to process in Input stream. Completing step.");

                returnValue = BatchConstants.STEP_COMPLETE;
            }
        }
    	catch (BatchContainerDataStreamException e)
    	{            	
    		throw new RuntimeException("failed to obtain batch input/output streams", e);
    	}        
        catch(Exception e) {
            throw new RuntimeException("Unexpected error in batch loop", e);
        }
        return returnValue;
    }

    protected Object processRecord(Object record) throws Exception
    {
    	String methodName = "processRecord";
    	if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
    	Object rec = null;
    	rec = batchRecordProcessor.processRecord(record);
    	if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
    	return rec;
    }

    /**
     * Called by LREE to set our properties as specified in xJCL.
     */
    public void setProperties(Properties props) {
        this.props = props;
        props.put("JobStepId", getJobID());
    }

   
    public String getJobID() {
        return  (String) getProperties().get(BatchConstants.JOB_ID); //JobStepContextMgr.getContext().getJobID();// return Integer.toString(hashCode());

    }
    

    public String getStepID() {
    	String jobID = (String) getProperties().get(BatchConstants.JOB_ID);
        String stepID = (String) getProperties().get(BatchConstants.STEP_ID);

        //System.out.println("SSA: GenericXDBatchStep.getStepID(): jobID= " + jobID + ", stepID= " + stepID);
        // System.out.println("getStepID jobID " + jobID + " stepID " + stepID);
    	// the bds logical name and the jobid/stepid are needed to retrieve the bds. 
        JobStepID jsId = new JobStepID( jobID, stepID );
        
        return  jsId.getJobstepid(); //JobStepContextMgr.getContext().getJobID();// return Integer.toString(hashCode());

    }
    
    protected Object loadClass(String className)
	{
		try
		{
			Object retval = Thread.currentThread().getContextClassLoader().loadClass(className).newInstance();
			return retval;
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}		
	}
}
