// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.steps.technologyadapters;

/**
 * This is an error tolerant batch step that relies on a ThresholdPolicy
 * to determine whether the number of errors that occurred during 
 * BatchRecordProcess.processRecord() is within a tolerable limit.
 */
import com.ibm.websphere.batch.BatchConstants;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.BatchDataStreamMgr;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStream;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataOutputStream;
import com.ibm.websphere.batch.devframework.thresholdpolicies.ThresholdPolicy;

public class ThresholdBatchStep extends GenericXDBatchStep {
	
	protected AbstractBatchDataOutputStream errorStream;
	protected ThresholdPolicy thresholdPolicy;
	protected ThreadLocal thresholdPolicyHandle;	
	protected int currentRecordCount;
	
	public void createJobStep() {
    	String methodName = "createJobStep";
        try {            	        	
        	
        	logger = new BDSFWLogger(this.getProperties());
        	perfAnalyzer = new PerformanceAnalyzer(CLASSNAME + "." + "processRecord", props);
        	isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
        	        	
        	if (logger.isDebugEnabled())
        		logger.debug("entering createJobStep()- jobId="+getJobID());

        	time = System.currentTimeMillis();
            
            inputStream = (AbstractBatchDataInputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.INPUT_STREAM_KEY, getStepID());
    		outputStream = (AbstractBatchDataOutputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.OUTPUT_STREAM_KEY, getStepID());
    		errorStream = (AbstractBatchDataOutputStream)BatchDataStreamMgr.getBatchDataStream(BDSFrameworkConstants.ERROR_STREAM_KEY, getStepID());
    		
    		if (inputStream == null)
    			logger.error("No Inputstream defined for the batch step!");
    		
    		if (outputStream == null)
    			logger.error("No Outputstream defined for the batch step!");
    		
    		if (errorStream == null)
    			logger.error("No Errorstream defined for the batch step!");
    		
    		try {
                
            	_header = inputStream.fetchHeader();
            	if (logger.isDebugEnabled())
            		logger.debug("GenericXDBatchStep: Got header" + _header);
            
            } catch (Exception e) {
            	// No header data
            	_header = null;
            	
            }
    		
    		String batchRecordProcessorImplClass = getRequiredProperty(BDSFrameworkConstants.BATCH_RECORD_PROCESSOR_KEY);
    		batchRecordProcessor = (BatchRecordProcessor)loadClass(batchRecordProcessorImplClass);
    		batchRecordProcessor.initialize(props);    		    	
    		
    		batchRecordProcessorHandle = new ThreadLocal();
    		batchRecordProcessorHandle.set(batchRecordProcessor);    		
    		
    		String thresholdPolicyImplClass = getRequiredProperty(BDSFrameworkConstants.THRESHOLD_POLICY_KEY);
    		thresholdPolicy = (ThresholdPolicy)loadClass(thresholdPolicyImplClass);
    		thresholdPolicy.initialize(props);   
    		
    		thresholdPolicyHandle = new ThreadLocal();
    		thresholdPolicyHandle.set(thresholdPolicy); 
    		
    		currentRecordCount = 0;
    		
        	if (logger.isDebugEnabled())
        		logger.debug("exiting createJobStep");

        }
        catch(Exception e) {
            throw new RuntimeException("Unexpected error in step setup", e);
        }  
    }
	
	/**
	 * Fetches a record from the input stream, passes it to the BatchRecordProcessor.
	 * processRecord. If the returned value is null it invokes the 
	 * ThresholdPolicy.invalidRecordEncountered method to determine if the number
	 * of errors occured so far is tolerable. If the ThresholdPolicy says it has
	 * exceeded the tolerable error threshold a checkpoint is forced and the job
	 * is cancelled.
	 * 
	 */
	
	 public int processJobStep() {    	    	
	    	String methodName = "processJobStep";
	    	int returnValue = -1;
	    	
	    	if (logger.isDebugEnabled())
	    		logger.debug("processJobStep");
	    	

	    	try
	    	{    		
	    		if (inputStream == null)
	    			logger.error("No Inputstream defined for the batch step!");
	    		
	    		if (outputStream == null)
	    			logger.error("No Outputstream defined for the batch step!");
	    		
	    		if (errorStream == null)
	    			logger.error("No error stream defined for the batch step!");
	    		
	    		if(_header != null && !_headerWritten ) {
	    			if (logger.isDebugEnabled()) 
	    				logger.debug("GenericXDBatchStep:processJobStep Writing header" + _header);
	    			outputStream.writeHeader(_header);
	    			_headerWritten = true;
	    		}
	    		
	            if (inputStream.hasNext()) {
	            	
	            	if (logger.isDebugEnabled())
	            		logger.debug("Inputstream has data to process.");

	                Object record = inputStream.read(); 
	            	
	                if (logger.isDebugEnabled())
	            		logger.debug("read record = " + record);
	                
	                
	                Object processedRecord = processRecord(record);
	                
	                if (processedRecord == null)
	                {
	                	logger.error("invalid record processed = " + record);
	                	errorStream.write(record);
	                	if (thresholdPolicy.invalidRecordEncountered(++currentRecordCount) == ThresholdPolicy.THRESHOLD_VIOLATION)
	                	{
	                		logger.error("Threshold policy has been violated. Gracefully stopping and placing job in restartable state.");
	                		returnValue = BatchConstants.STEP_CONTINUE_FORCE_CHECKPOINT_BEFORE_PROCESSING_CANCEL;
	                	}
	                	
	                }
	                else
	                {           
	                	if (logger.isDebugEnabled())
	                		logger.debug("processed record = " + processedRecord);
	               		outputStream.write(processedRecord);
	               		returnValue = BatchConstants.STEP_CONTINUE;
	                }
	            }
	            else {
	            	
	            	if (logger.isDebugEnabled())
	            		logger.debug("No more records to process in Input stream. Completing step.");

	                returnValue = BatchConstants.STEP_COMPLETE;
	            }
	        }
	    	catch (BatchContainerDataStreamException e)
	    	{            	
	    		throw new RuntimeException("failed to obtain batch input/output streams", e);
	    	}        
	        catch(Exception e) {
	            throw new RuntimeException("Unexpected error in batch loop", e);
	        }
	       
	        return returnValue;
	        
	    }
	 
	 /**
	  * Calls BatchRecordProcessor.completeProcessing() to determine
	  * the return value of the step.
	  */
	 public int destroyJobStep() { 
	    	
	    	if (logger.isDebugEnabled())
	    		logger.debug("entering destroyJobStep");
	    	int returnValue = -1;
	    	batchRecordProcessor = (BatchRecordProcessor) batchRecordProcessorHandle.get();
	    	returnValue = batchRecordProcessor.completeProcessing();
	    	
	    	thresholdPolicy = (ThresholdPolicy) thresholdPolicyHandle.get();
	    	thresholdPolicy.destroy();
	    	
	    	if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
	                
	        logger.info("GenericXDBatchStep.destroyStep()- Total Execution Time: " + (System.currentTimeMillis() - time));
	        
	    	if (logger.isDebugEnabled())
	    		logger.debug( "destroyJobStep");

	        return returnValue;
	    }

}
