// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
package com.ibm.websphere.batch.devframework.test.sample.steps;

import java.util.Properties;

import com.ibm.websphere.batch.devframework.steps.technologyadapters.BatchRecordProcessor;

public class InfrastructureVerificationTest implements BatchRecordProcessor {

	
	//	3 behavioral properties:
	protected String CPU_BOUND_KEY = "CPU_BOUND";  // value is number of calculation iterations
	protected String MEM_BOUND_KEY = "MEM_BOUND";  // value is (in bytes) amount of memory in KB to allocate
	protected String FORCE_FAILURE_KEY = "FORCE_FAILURE";  // value is number of iterations before failing
	protected String FORCE_MEM_LEAK_KEY = "FORCE_MEM_LEAK"; // value is (in bytes) amount of memory to leak on each iteration.
	protected String MAX_MEM_LEAK_KEY = "MAX_MEM_LEAK"; // value is (in bytes) max amount of memory to leak for the step.
	protected String FORCE_EXIT_KEY = "FORCE_EXIT"; // value is number of iterations before system.exit
	protected String CHAINED_RECORD_PROCESSOR_KEY = "CHAINED_PROCESSOR"; // value is a second record processor that does some business logic. 
	
	protected BatchRecordProcessor chainedRecordProcessor = null;
	protected boolean chainedProcessorDefined = false;
	
	protected boolean CPU_BOUND = false;
	protected int cpuIterations = 10;
	
	protected boolean MEM_BOUND = false;
	protected int memAllocation = 500;
	protected char[] consumed_memory;
	
	protected boolean FORCE_FAILURE = false;
	protected int failureIterations = 10;
	protected int failureCount = 0;
	
	protected boolean FORCE_MEM_LEAK = false;
	protected int memLeakAllocation = 100;
	protected int leakIteration = 0;
	protected char[] leakedMemory;
	
	protected boolean MAX_MEM_LEAK = false;
	protected int maxMemLeak = 5000;
	protected boolean maxMemLeakReached = false;
	
	protected boolean FORCE_EXIT = false;
	protected int exitIterations = 10;
	protected int exitCount = 0;
	
	public void initialize(Properties props) 
	{								
		
		String chainedProcessorImplClass = props.getProperty(CHAINED_RECORD_PROCESSOR_KEY);
		if (chainedProcessorImplClass != null)
		{			
			try
			{
				chainedRecordProcessor = (BatchRecordProcessor) loadClass(chainedProcessorImplClass);
				chainedRecordProcessor.initialize(props);
				chainedProcessorDefined = true;
			}
			catch (Throwable t)
			{
				throw new RuntimeException(t);
			}			
		}
		
		String cpuProp = props.getProperty(CPU_BOUND_KEY);
		if (cpuProp != null)
		{
			CPU_BOUND = true;
			cpuIterations = Integer.valueOf(cpuProp).intValue();			
		}
		
		String memProp = props.getProperty(MEM_BOUND_KEY);
		if (memProp != null)
		{
			MEM_BOUND = true;
			memAllocation = Integer.valueOf(memProp).intValue();			
		}
		
		String failureProp = props.getProperty(FORCE_FAILURE_KEY);
		if (failureProp != null)
		{
			FORCE_FAILURE = true;
			failureIterations = Integer.valueOf(failureProp).intValue();			
		}
		
		String leakProp = props.getProperty(FORCE_MEM_LEAK_KEY);
		if (leakProp != null)
		{
			FORCE_MEM_LEAK = true;
			memLeakAllocation = Integer.valueOf(leakProp).intValue();			
		}
		
		String maxLeakProp = props.getProperty(MAX_MEM_LEAK_KEY);
		if (maxLeakProp != null)
		{
			MAX_MEM_LEAK = true;
			maxMemLeak = Integer.valueOf(maxLeakProp).intValue();			
		}
		
		String exitProp = props.getProperty(FORCE_EXIT_KEY);
		if (exitProp != null)
		{
			FORCE_EXIT = true;
			exitIterations = Integer.valueOf(exitProp).intValue();			
		}

		wasteMemory();
	}
	
	protected void forceMemLeak()
	{
		if (FORCE_MEM_LEAK)
		{
			if (!maxMemLeakReached)
			{
				leakIteration++;
				int memToLeak = memLeakAllocation * leakIteration;
				leakedMemory = new char[memToLeak];
				
				if (memToLeak >= maxMemLeak)
					maxMemLeakReached = true;
			}
		}
	}

	protected void forceFailure()
	{
		if (FORCE_FAILURE)
		{			
			failureCount++;
			if ((failureCount % failureIterations) == 0)
				throw new RuntimeException("Forced Failure!");
		}
	}
	
	protected void forceExit()
	{
		if (FORCE_EXIT)
		{			
			exitCount++;
			if ((exitCount % exitIterations) == 0)
				System.exit(7);
		}
	}
	
	protected void wasteMemory()
	{
		if (MEM_BOUND)
		{
			consumed_memory = new char[memAllocation];
		}
	}
	
	protected void wasteCPU()
	{
		if (CPU_BOUND)
		{
			for (int i = 0; i < cpuIterations; i++)
			{
				double ret = Math.exp(Math.log(Math.exp(Math.log(Math.exp(Math.log(Math.exp(Math.log(1000000))))))));				
			}
		}
	}
	
	public Object processRecord(Object record) throws Exception {
		forceMemLeak();
		wasteCPU();
		forceFailure();
		forceExit();
		
		if (chainedProcessorDefined)
		{
			record = chainedRecordProcessor.processRecord(record);
		}
		
		return record;
	}
	
	public int completeProcessing() 
	{
		int returnCode = 0;
		if (chainedProcessorDefined)
		{
			returnCode = chainedRecordProcessor.completeProcessing();
		}
		return returnCode;
	}
	
	protected Object loadClass(String className)
	{
		try
		{
			Object retval = Thread.currentThread().getContextClassLoader().loadClass(className).newInstance();
			return retval;
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
			
	}

}
