// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

/**
 * An exemplary implementation of the ThresholdPolicy interface that
 * calculates the % of error records to the total processed if the 
 * result is greater then returns a THRESHOLD_VIOLATION
 */
package com.ibm.websphere.batch.devframework.thresholdpolicies;

import java.util.Properties;

import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;

public class PercentageBasedThresholdPolicy implements ThresholdPolicy {

	protected int currentInvalidRecordsProcessed;
	
	protected String defaultThresholdPercentage = ".1"; // 10%
	protected double thresholdPercentage = .1;
	protected String defaultMinSampleData = "20";
	protected int minSampleData = 20;
	protected BDSFWLogger logger;
	
	public void initialize(Properties props) {
		
		logger = new BDSFWLogger(props);
		
		if (logger.isDebugEnabled())
    		logger.debug("entering ThresholdPolicy.intialize");
		
		String thresholdPercentageValue = props.getProperty(BDSFrameworkConstants.THRESHOLD_PERCENTAGE_KEY, defaultThresholdPercentage);
		String minSampleDataStr = props.getProperty(BDSFrameworkConstants.MINIMUM_THRESHOLD_SAMPLESIZE_KEY, defaultMinSampleData);
		
		try 
		{
			thresholdPercentage = Double.valueOf(thresholdPercentageValue).doubleValue();
			minSampleData = Integer.valueOf(minSampleDataStr).intValue();
		}
		catch (Exception e)
		{
			logger.error("Error setting threshold percentage. Defaulting to " + defaultThresholdPercentage + ". Exception:" + e);
		}
		
		if (logger.isDebugEnabled())
		{
			logger.debug("thresholdPercentage is set to: " + thresholdPercentage);
		}
		currentInvalidRecordsProcessed = 0;
		
		if (logger.isDebugEnabled())
    		logger.debug("exiting ThresholdPolicy.intialize");

	}

	/**
	 * Increments the currentInvalidRecordsProcessed count then checks whether it is greater
	 * than the minimum Sample data threshold, if it is then it determines the % of 
	 * the number of invalid records at this moment to the total record count and checks whether
	 * it exceeds the threshold. If so then returns a threshold violation
	 * 
	 */
	public int invalidRecordEncountered(int currentRecordCount) {
		
		int retval = ThresholdPolicy.NO_THRESHOLD_VIOLATION;
		
		currentInvalidRecordsProcessed++;
		
		if (logger.isDebugEnabled())
    		logger.debug("entering ThresholdPolicy.invalidRecordEncountered. CurrentRecordCount=" + currentRecordCount + ", currentInvalidRecordsProcessed=" + currentInvalidRecordsProcessed);
		
		if (currentRecordCount > minSampleData)
		{
			if ((currentInvalidRecordsProcessed / currentRecordCount) > thresholdPercentage)
				retval = ThresholdPolicy.THRESHOLD_VIOLATION;
		}
		
		if (logger.isDebugEnabled())
    		logger.debug("exiting ThresholdPolicy.invalidRecordEncountered. Retval = " + retval);
		
		return retval;
	}
	
	/**
	 * No Op
	 */
	public void destroy() {
		if (logger.isDebugEnabled())
    		logger.debug("entering ThresholdPolicy.destroy");
		
		// no op.
		
		if (logger.isDebugEnabled())
    		logger.debug("exiting ThresholdPolicy.destroy");
	}

}
