// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

package com.ibm.websphere.batch.devframework.thresholdpolicies;

/**
 * An exemplary implementation of the ThresholdPolicy interface that keeps track
 * of the number of error records processed and returns a policy violation if it 
 * exceeds the specified threshold
 */
import java.util.Properties;

import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;

public class RecordBasedThresholdPolicy implements ThresholdPolicy {

	protected int currentInvalidRecordsProcessed;
	
	protected String defaultThreshold = "100"; // 10%
	protected int threshold = 100;
	String defaultMinSampleData = "20";
	protected int minSampleData = 20;
	protected BDSFWLogger logger;
	
	public void initialize(Properties props) {
		
		logger = new BDSFWLogger(props);
		
		if (logger.isDebugEnabled())
    		logger.debug("entering ThresholdPolicy.intialize");
		
		String thresholdPercentageValue = props.getProperty(BDSFrameworkConstants.THRESHOLD_RECORD_COUNT, defaultThreshold);
		String minSampleDataStr = props.getProperty(BDSFrameworkConstants.MINIMUM_THRESHOLD_SAMPLESIZE_KEY, defaultMinSampleData);
		try 
		{
			threshold = Integer.valueOf(thresholdPercentageValue).intValue();
			minSampleData = Integer.valueOf(minSampleDataStr).intValue();
		}
		catch (Exception e)
		{
			logger.error("Error setting threshold. Defaulting to " + defaultThreshold + ". Exception:" + e);
		}
		
		if (logger.isDebugEnabled())
		{
			logger.debug("threshold is set to: " + threshold);
		}
		currentInvalidRecordsProcessed = 0;
		
		if (logger.isDebugEnabled())
    		logger.debug("exiting ThresholdPolicy.intialize");

	}

	/**
	 * Increments the number of invalid records encountered, if this count is greater than
	 * the minimum sample threshold then checks if it exceeds the invalid records threshold
	 * if it does returns a threshold violation
	 */
	public int invalidRecordEncountered(int currentRecordCount) {
		
		int retval = ThresholdPolicy.NO_THRESHOLD_VIOLATION;
		
		currentInvalidRecordsProcessed++;
		
		if (logger.isDebugEnabled())
    		logger.debug("entering ThresholdPolicy.invalidRecordEncountered. CurrentRecordCount=" + currentRecordCount + ", currentInvalidRecordsProcessed=" + currentInvalidRecordsProcessed);
		
		if (currentRecordCount > minSampleData)
		{
			if (currentInvalidRecordsProcessed > threshold)
				retval = ThresholdPolicy.THRESHOLD_VIOLATION;
		}
		
		if (logger.isDebugEnabled())
    		logger.debug("exiting ThresholdPolicy.invalidRecordEncountered. Retval = " + retval);
		
		return retval;
	}
	
	/**
	 * no op
	 */
	public void destroy() {
		if (logger.isDebugEnabled())
    		logger.debug("entering ThresholdPolicy.destroy");
		
		// no op.
		
		if (logger.isDebugEnabled())
    		logger.debug("exiting ThresholdPolicy.destroy");
	}

}
